from flask_babel import lazy_gettext as _
from hiddifypanel.hutils import LazyLoader
requests=LazyLoader("requests")
from packaging.version import Version
import re
import sys

from hiddifypanel.models.config import hconfig, ConfigEnum
from hiddifypanel import __version__ as current_version
from hiddifypanel.cache import cache


to_gig_d = 1000 * 1000 * 1000


def error(str):
    print(str, file=sys.stderr)


@cache.cache(ttl=60000)
def get_latest_release_url(repo):
    latest_url = requests.get(f'{repo}/releases/latest').url.strip()
    version = latest_url.split('tag/')[1].strip()
    return (latest_url, version)


@cache.cache(ttl=600)
def get_latest_release_version(repo_name):
    try:
        url = f"https://github.com/hiddify/{repo_name}/releases/latest"
        response = requests.head(url, allow_redirects=False)

        location_header = response.headers.get("Location")
        if location_header:
            version = re.search(r"/([^/]+)/?$", location_header)
            if version:
                ver = version.group(1).replace('v', '')
                if ver == "latest":
                    return get_latest_release_version(repo_name.replace("-", ""))
                return ver
    except Exception:
        return None

    return None


def is_panel_outdated() -> bool:
    # TODO: handle beta and develop version too
    pm = hconfig(ConfigEnum.package_mode)
    try:
        if pm == 'release':
            if latest_v := get_latest_release_version('hiddifypanel'):
                if compare_versions(latest_v, current_version) == 1:
                    return True
    except:
        pass
    return False


def compare_versions(version_1: str, version_2: str) -> int:
    """
    Compare two version strings and return an integer based on their relative order.
    Returns:
        int:

        - 1 if version_1 is greater than version_2.
        - 0 if version_1 is equal to version_2.
        - -1 if version_1 is less than version_2.

    Examples:
        >>> compare_versions("10.20.4", "10.20.4")
        0
        >>> compare_versions("10.20.4", "10.20.2")
        1
        >>> compare_versions("10.20.2", "10.20.4")
        -1
    """
    v1 = Version(version_1)
    v2 = Version(version_2)

    if v1 > v2:
        return 1  # version_1 is greater
    elif v2 > v1:
        return -1  # version_2 is greater
    else:
        return 0  # versions are equal



import psutil
import os


def get_folder_size(folder_path: str) -> int:
    total_size = 0
    try:
        for dirpath, dirnames, filenames in os.walk(folder_path):
            for file in filenames:
                file_path = os.path.join(dirpath, file)
                try:
                    total_size += os.path.getsize(file_path)
                except BaseException:
                    pass
    except BaseException:
        pass
    return total_size


def top_processes() -> dict:
    # Get the process information
    processes = [p for p in psutil.process_iter(['pid', 'name', 'username', 'cpu_percent', 'memory_info']) if p.info['name'] != '']
    num_cores = psutil.cpu_count()
    # Calculate memory usage, RAM usage, and CPU usage for each process
    memory_usage = {}
    ram_usage = {}
    cpu_usage = {}
    for p in processes:
        name = p.info['name']
        if p.info['username']=="hiddify-panel":
            name = "Hiddify"
        # mem_info = p.info['memory_full_info']
        # if mem_info is None:
        #     continue
        # mem_usage = mem_info.uss
        mem_usage = p.info['memory_info'].rss
        cpu_percent = p.info['cpu_percent'] / num_cores
        if name in memory_usage:
            memory_usage[name] += mem_usage / (1024 ** 3)
            ram_usage[name] += mem_usage / (1024 ** 3)
            cpu_usage[name] += cpu_percent
        else:
            memory_usage[name] = mem_usage / (1024 ** 3)
            ram_usage[name] = mem_usage / (1024 ** 3)
            cpu_usage[name] = cpu_percent

    while len(cpu_usage) < 5:
        cpu_usage[" " * len(cpu_usage)] = 0
    while len(ram_usage) < 5:
        ram_usage[" " * len(ram_usage)] = 0
    while len(memory_usage) < 5:
        memory_usage[" " * len(memory_usage)] = 0
    # Sort the processes by memory usage, RAM usage, and CPU usage
    top_memory = sorted(memory_usage.items(), key=lambda x: x[1], reverse=True)[:5]
    top_ram = sorted(ram_usage.items(), key=lambda x: x[1], reverse=True)[:5]
    top_cpu = sorted(cpu_usage.items(), key=lambda x: x[1], reverse=True)[:5]

    # Return the top processes for memory usage, RAM usage, and CPU usage
    return {
        "memory": top_memory,
        "ram": top_ram,
        "cpu": top_cpu
    }


def system_stats() -> dict:
    # CPU usage
    cpu_percent = psutil.cpu_percent(interval=1)

    # RAM usage
    ram_stats = psutil.virtual_memory()
    ram_used = ram_stats.used / 1024**3
    ram_total = ram_stats.total / 1024**3

    # Disk usage (in GB)
    disk_stats = psutil.disk_usage('/')
    disk_used = disk_stats.used / 1024**3
    disk_total = disk_stats.total / 1024**3

    hiddify_used = get_folder_size('/opt/hiddify-manager/') / 1024**3

    # Network usage
    net_stats = psutil.net_io_counters()
    bytes_sent_cumulative = net_stats.bytes_sent
    bytes_recv_cumulative = net_stats.bytes_recv
    bytes_sent = net_stats.bytes_sent - getattr(system_stats, 'prev_bytes_sent', 0)
    bytes_recv = net_stats.bytes_recv - getattr(system_stats, 'prev_bytes_recv', 0)
    system_stats.prev_bytes_sent = net_stats.bytes_sent
    system_stats.prev_bytes_recv = net_stats.bytes_recv

    # Total connections and unique IPs
    connections = psutil.net_connections()
    total_connections = len(connections)
    unique_ips = set([conn.raddr.ip for conn in connections if conn.status == 'ESTABLISHED' and conn.raddr])
    total_unique_ips = len(unique_ips)

    # Load average
    num_cpus = psutil.cpu_count()
    load_avg = [avg / num_cpus for avg in os.getloadavg()]
    # Return the system information
    return {
        "cpu_percent": cpu_percent / num_cpus,
        "ram_used": ram_used,
        "ram_total": ram_total,
        "disk_used": disk_used,
        "disk_total": disk_total,
        "hiddify_used": hiddify_used,
        "bytes_sent": bytes_sent,
        "bytes_recv": bytes_recv,
        "bytes_sent_cumulative": bytes_sent_cumulative,
        "bytes_recv_cumulative": bytes_recv_cumulative,
        "net_sent_cumulative_GB": bytes_sent_cumulative / 1024**3,
        "net_total_cumulative_GB": (bytes_sent_cumulative + bytes_recv_cumulative) / 1024**3,
        "total_connections": total_connections,
        "total_unique_ips": total_unique_ips,
        "load_avg_1min": load_avg[0],
        "load_avg_5min": load_avg[1],
        "load_avg_15min": load_avg[2],
        'num_cpus': num_cpus
    }



from .role import Role, AccountType
from .child import Child, ChildMode
from .config_enum import ConfigCategory, ConfigEnum, Lang, ApplyMode, PanelMode, LogLevel,MieruHandshake,MieruMultiplexing
from .config import StrConfig, BoolConfig, get_hconfigs, hconfig, set_hconfig, add_or_update_config, bulk_register_configs, get_hconfigs_childs

# from .parent_domain import ParentDomain
from .domain import Domain, DomainType, ShowDomain
from .proxy import Proxy, ProxyL3, ProxyCDN, ProxyProto, ProxyTransport
from .user import User, UserMode, UserDetail, ONE_GIG
from .admin import AdminUser, AdminMode
from .usage import DailyUsage
from .base_account import BaseAccount
# from .report import Report, ReportDetail



from flask_babel import gettext as _
from flask_babel import lazy_gettext as __
# from flask_babel import ngettext as n_
from flask_babel import LazyString
from flask_admin.babel import gettext as gettext_old
from flask_admin.babel import ngettext as ngettext_old
from flask_admin.babel import lazy_gettext as lazy_gettext_old
import flask_admin.babel


def gettext(string, **variables):

    tt = _(string, **variables)
    # print("=--", string, tt)
    if tt == string:
        return gettext_old(string, **variables)
    return tt


# def ngettext(*args, **kwargs):
#     tt = n_(*args, **kwargs)
#     if tt == string:
#         return ngettext_old(string, **variables)
#     return tt


def lazy_gettext(string, **variables):
    return LazyString(gettext, string, **variables)


flask_admin.babel.gettext = gettext
# flask_admin.babel.ngettext = ngettext
flask_admin.babel.lazy_gettext = lazy_gettext



from apiflask import abort
from hiddifypanel.database import db
from flask import current_app as app
from flask.views import MethodView
from loguru import logger

from hiddifypanel.models import *
from hiddifypanel.auth import login_required

from .schema import RegisterInputSchema, RegisterOutputSchema


class RegisterApi(MethodView):
    decorators = [login_required({Role.super_admin})]

    @app.input(RegisterInputSchema, arg_name='data')  # type: ignore
    @app.output(RegisterOutputSchema)  # type: ignore
    def put(self, data):
        from hiddifypanel import hutils
        logger.info("Register child panel with unique_id: {}", data['unique_id'])
        if hutils.node.is_child():
            logger.error("The panel in child, not a parent nor standalone")
            abort(400, 'The panel in child, not a parent nor standalone')

        unique_id = data['unique_id']
        name = data['name']
        mode = data['mode']

        child = Child.query.filter(Child.unique_id == unique_id).first()
        if not child:
            logger.info("Adding new child with unique_id: {}", unique_id)
            child = Child(unique_id=unique_id, name=name, mode=mode)
            db.session.add(child)  # type: ignore
            db.session.commit()  # type: ignore
            child = Child.query.filter(Child.unique_id == unique_id).first()

        try:
            # add data
            logger.info("Adding admin users...")
            AdminUser.bulk_register(data['panel_data']['admin_users'], commit=False)
            logger.info("Adding users...")
            User.bulk_register(data['panel_data']['users'], commit=False)
            logger.info("Adding domains...")
            Domain.bulk_register(data['panel_data']['domains'], commit=False, force_child_unique_id=child.unique_id)
            logger.info("Adding hconfigs...")
            bulk_register_configs(data['panel_data']['hconfigs'], commit=False, froce_child_unique_id=child.unique_id)
            logger.info("Adding proxies...")
            Proxy.bulk_register(data['panel_data']['proxies'], commit=False, force_child_unique_id=child.unique_id)
            db.session.commit()  # type: ignore
        except Exception as err:
            with logger.contextualize(error=err):
                logger.error("Error while registering data")
            abort(400, str(err))

        if not hutils.node.is_parent():
            logger.info("Setting panel to parent mode")
            set_hconfig(ConfigEnum.panel_mode, PanelMode.parent)  # type: ignore

        logger.info("Returning register output")
        res = RegisterOutputSchema()
        res.users = [u.to_schema() for u in User.query.all()]  # type: ignore
        res.admin_users = [a.to_schema() for a in AdminUser.query.all()]  # type: ignore
        res.parent_unique_id = hconfig(ConfigEnum.unique_id)

        return res



from flask import g, send_from_directory
from hiddifypanel.hutils.flask import hurl_for


def send_static(path):
    return send_from_directory('static/assets', path)


def init_app(app):
    @app.url_defaults
    def add_asset(endpoint, values):
        def get_asset(path):
            return hurl_for('asset', proxy_path=g.proxy_path, path=path)
        g.asset_url = get_asset

    app.add_url_rule('/<proxy_path>/asset/<path:path>',
                     'asset', view_func=send_static)
    pass



import random
import string
from hiddifypanel.models import hconfig, ConfigEnum


def get_random_string(min_: int = 10, max_: int = 30) -> str:
    # With combination of lower and upper case
    length = random.randint(min_, max_)
    characters = string.ascii_letters + string.digits
    result_str = ''.join(random.choice(characters) for i in range(length))
    return result_str


def get_random_password(length: int = 16) -> str:
    '''Retunrns a random password with fixed length'''
    characters = string.ascii_letters + string.digits  # + '-'
    while True:
        passwd = ''.join(random.choice(characters) for i in range(length))
        if (any(c.islower() for c in passwd) and any(c.isupper() for c in passwd) and sum(c.isdigit() for c in passwd) > 1):
            return passwd


def __is_port_in_range(port, start_port: int | str | None, count: int):
    if start_port is None:
        return False
    start_port = int(start_port)
    if port < start_port:
        return False

    if port > start_port + count:
        return False
    return True


def __is_in_used_port(port):
    for k in ConfigEnum:
        if "port" in k:
                for p in (hconfig(k) or "").split(","):
                    if p and __is_port_in_range(port, p, 100):
                        return True
    
    for p in [443, 80, 9000, 10085, 10086]:
        if __is_port_in_range(port, p, 100):
            return True
    return False


def get_random_unused_port():
    port = random.randint(11000, 60000)
    while __is_in_used_port(port):
        port = random.randint(11000, 60000)
    return port


def random_case(string):
    return ''.join(random.choice((str.upper, str.lower))(c) for c in string)



from loguru import logger
import socket
from flask_babel import gettext as _
from strenum import StrEnum

from hiddifypanel.models import AdminUser, User, hconfig, ConfigEnum, ChildMode, Domain, Proxy, StrConfig, BoolConfig, Child, ChildMode
from hiddifypanel import hutils
from hiddifypanel.panel import hiddify
from hiddifypanel.panel import usage
from hiddifypanel.database import db
from hiddifypanel.cache import cache

# import schmeas
from hiddifypanel.panel.commercial.restapi.v2.parent.schema import *
from hiddifypanel.panel.commercial.restapi.v2.child.schema import *

from .api_client import NodeApiClient, NodeApiErrorSchema
# region private


def __get_register_data_for_api(name: str, mode: ChildMode) -> RegisterInputSchema:

    register_data = RegisterInputSchema()
    register_data.unique_id = hconfig(ConfigEnum.unique_id)
    register_data.name = name  # type: ignore
    register_data.mode = mode  # type: ignore

    panel_data = RegisterDataSchema()  # type: ignore
    panel_data.admin_users = [admin_user.to_schema() for admin_user in AdminUser.query.all()]  # type: ignore
    panel_data.users = [user.to_schema() for user in User.query.all()]  # type: ignore
    panel_data.domains = [domain.to_schema() for domain in Domain.query.all()]  # type: ignore
    panel_data.proxies = [proxy.to_schema() for proxy in Proxy.query.all()]  # type: ignore
    panel_data.hconfigs = [*[u.to_schema() for u in StrConfig.query.all()], *[u.to_schema() for u in BoolConfig.query.all()]]  # type: ignore
    register_data.panel_data = panel_data

    return register_data


class SyncFields(StrEnum):
    domains = 'domains'
    proxies = 'proxies'
    hconfigs = 'hconfigs'


def __get_sync_data_for_api(*fields: SyncFields) -> SyncInputSchema:
    sync_data = SyncInputSchema()
    if len(fields) == 0:
        sync_data.domains = [domain.to_schema() for domain in Domain.query.all()]  # type: ignore
        sync_data.proxies = [proxy.to_schema() for proxy in Proxy.query.all()]  # type: ignore
        sync_data.hconfigs = [*[u.to_schema() for u in StrConfig.query.all()], *[u.to_schema() for u in BoolConfig.query.all()]]  # type: ignore
    else:
        for f in fields:
            match f:
                case SyncFields.domains:
                    sync_data.domains = [domain.to_schema() for domain in Domain.query.all()]  # type: ignore
                case SyncFields.proxies:
                    sync_data.proxies = [proxy.to_schema() for proxy in Proxy.query.all()]  # type: ignore
                case SyncFields.hconfigs:
                    sync_data.hconfigs = [*[u.to_schema() for u in StrConfig.query.all()], *[u.to_schema() for u in BoolConfig.query.all()]]  # type: ignore

    return sync_data


def __get_parent_panel_url() -> str:
    url = 'https://' + f"{hconfig(ConfigEnum.parent_domain).removesuffix('/')}/{hconfig(ConfigEnum.parent_admin_proxy_path).removesuffix('/')}"
    return url

# endregion


def is_registered() -> bool:
    '''Checks if the current parent registered as a child'''
    try:
        logger.debug("Checking if current panel is registered with parent")
        base_url = __get_parent_panel_url()
        if not base_url:
            return False
        payload = ChildStatusInputSchema()
        payload.child_unique_id = hconfig(ConfigEnum.unique_id)

        res = NodeApiClient(base_url).post('/api/v2/parent/status/', payload, ChildStatusOutputSchema)
        if isinstance(res, NodeApiErrorSchema):
            logger.error(f"Error while checking if current panel is registered with parent: {res.msg}")
            return False

        if res['existance']:
            return True
        return False
    except Exception as e:
        logger.error(f"Error while checking if current panel is registered with parent")
        logger.exception(e)
        return False


def register_to_parent(name: str, apikey: str, mode: ChildMode = ChildMode.remote) -> bool:
    # get parent link its format is "https://panel.hiddify.com/<admin_proxy_path>/"
    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Parent url is empty")
        return False

    payload = __get_register_data_for_api(name, mode)
    res = NodeApiClient(p_url, apikey).put('/api/v2/parent/register/', payload, RegisterOutputSchema)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while registering to parent: {res.msg}")
        return False

    # TODO: change the bulk_register and such methods to accept models instead of dict
    AdminUser.bulk_register(res['admin_users'], commit=False)
    User.bulk_register(res['users'], commit=False)

    # add new child as parent
    db.session.add(  # type: ignore
        Child(unique_id=res['parent_unique_id'], name=socket.gethostname() or res['parent_unique_id'], mode=ChildMode.parent)
    )

    db.session.commit()  # type: ignore

    logger.success("Successfully registered to parent")
    cache.invalidate_all_cached_functions()
    return True


def sync_with_parent(*fields: SyncFields) -> bool:
    # sync usage first
    if not sync_users_usage_with_parent():
        logger.error("Error while syncing with parent: Failed to sync users usage")
        return False

    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Error while syncing with parent: Parent url is empty")
        return False
    payload = __get_sync_data_for_api(*fields)
    res = NodeApiClient(p_url).put('/api/v2/parent/sync/', payload, SyncOutputSchema)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while syncing with parent: {res.msg}")
        return False
    AdminUser.bulk_register(res['admin_users'], commit=False, remove=True)
    User.bulk_register(res['users'], commit=False, remove=True)
    db.session.commit()  # type: ignore
    logger.success("Successfully synced with parent")
    cache.invalidate_all_cached_functions()
    return True


def sync_users_usage_with_parent() -> bool:
    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Parent url is empty")
        return False

    payload = hutils.node.get_users_usage_data_for_api()
    if payload:
        res = NodeApiClient(p_url).put('/api/v2/parent/usage/', payload, UsageInputOutputSchema)  # type: ignore
        if isinstance(res, NodeApiErrorSchema):
            logger.error(f"Error while syncing users usage with parent: {res.msg}")
            return False

        # parse usages data
        res = hutils.node.convert_usage_api_response_to_dict(res)  # type: ignore
        usage.add_users_usage_uuid(res, hiddify.get_child(None), True)
        logger.success(f"Successfully synced users usage with parent: {res}")

    return True



from urllib.parse import urlparse
from flask import g, request
from flask import current_app as app
from flask.views import MethodView
from hiddifypanel.auth import login_required
from hiddifypanel.models import Proxy, Role, ConfigEnum, hconfig
from apiflask import Schema
from apiflask.fields import String
from hiddifypanel.panel.user.user import get_common_data
from hiddifypanel import hutils
from apiflask import fields

class ConfigSchema(Schema):
    name = String(required=True)
    domain = String(required=True)
    link = String(required=True)
    protocol = String(required=True)
    transport = String(required=True)
    security = String(required=True)
    type = String(required=True)


class AllConfigsAPI(MethodView):
    decorators = [login_required({Role.user})]

    @app.output(list[ConfigSchema])  # type: ignore
    def get(self):
        def create_item(name, domain, type, protocol, transport, security, link):
            dto = ConfigSchema()
            dto.name = name
            dto.type = type
            dto.domain = domain
            dto.protocol = protocol
            dto.transport = transport
            dto.security = security
            dto.link = link
            return dto

        items = []
        base_url = f"https://{urlparse(request.base_url).hostname}/{g.proxy_path}/{g.account.uuid}/"
        c = get_common_data(g.account.uuid, 'new')
        config_name = hutils.encode.url_encode(c['user'].name)

        # Add Auto
        items.append(
            create_item(
                "Auto", "ALL", "", "", "", "",
                # f"{base_url}sub/?asn={c['asn']}"
                f"{base_url}auto/?asn={c['asn']}#{config_name}"
            )
        )

        if hconfig(ConfigEnum.sub_full_singbox_enable):
            # Add Full Singbox
            items.append(
                create_item(
                    "Full Singbox", "ALL", "", "", "", "",
                    # f"{base_url}full-singbox.json?asn={c['asn']}"
                    f"{base_url}singbox/?asn={c['asn']}#{config_name}"
                )
            )

        if hconfig(ConfigEnum.sub_full_xray_json_enable):
            # Add Full Xray
            items.append(
                create_item(
                    "Full Xray", "ALL", "", "", "", "",
                    f"{base_url}xray/#{config_name}"
                )
            )

        if hconfig(ConfigEnum.sub_full_links_enable):
            # Add Subscription link
            items.append(
                create_item(
                    "Subscription link", "ALL", "", "", "", "",
                    # f"{base_url}all.txt?name={c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}&asn={c['asn']}&mode={c['mode']}"
                    f"{base_url}sub/?asn={c['asn']}#{config_name}"
                )
            )

        if hconfig(ConfigEnum.sub_full_links_b64_enable):
            # Add Subscription link base64
            items.append(
                create_item(
                    "Subscription link b64", "ALL", "", "", "", "",
                    # f"{base_url}all.txt?name=new_link_{c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}-{c['mode']}&asn={c['asn']}&mode={c['mode']}&base64=True"
                    f"{base_url}sub64/?asn={c['asn']}#{config_name}"
                )
            )
        if hconfig(ConfigEnum.sub_full_clash_meta_enable):
            # Add Clash Meta
            items.append(
                create_item(
                    "Clash Meta", "ALL", "", "", "", "",
                    # f"clashmeta://install-config?url={base_url}clash/meta/all.yml&name=mnormal_{c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}-{c['mode']}&asn={c['asn']}&mode={c['mode']}"
                    f"{base_url}clashmeta/?asn={c['asn']}#{config_name}"
                )
            )

        if hconfig(ConfigEnum.sub_full_clash_enable):
            # Add Clash
            items.append(
                create_item(
                    "Clash", "ALL", "Except VLess", "", "", "",
                    # f"clash://install-config?url={base_url}clash/all.yml&name=new_normal_{c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}-{c['mode']}&asn={c['asn']}&mode={c['mode']}"
                    f"{base_url}clash/?asn={c['asn']}#{config_name}"
                )
            )

        if hconfig(ConfigEnum.wireguard_enable):
            items.append(
                create_item(
                    "Wireguard", "Wireguard", "", "", "", "",
                    # f"{base_url}singbox.json?name={c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}&asn={c['asn']}&mode={c['mode']}"
                    f"{base_url}wireguard/#{config_name}"
                )
            )
            # Add Singbox: SSh
        if hconfig(ConfigEnum.sub_singbox_ssh_enable) and hconfig(ConfigEnum.ssh_server_enable):
            items.append(
                create_item(
                    "Singbox: SSH", "SSH", "", "", "", "",
                    # f"{base_url}singbox.json?name={c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}&asn={c['asn']}&mode={c['mode']}"
                    f"{base_url}singbox-ssh/?asn={c['asn']}#{config_name}"
                )
            )

        for pinfo in hutils.proxy.get_valid_proxies(c['domains']):
            items.append(
                create_item(
                    pinfo["name"].replace("_", " "),
                    f"{pinfo['mode']}",
                    pinfo['server'],
                    pinfo['proto'],
                    pinfo['transport'],
                    pinfo['l3'],
                    f"{hutils.proxy.xray.to_link(pinfo)}"
                )
            )

        return items



from flask import current_app as app
from flask import g
from flask.views import MethodView
from apiflask import abort
from hiddifypanel.auth import login_required
from hiddifypanel.models import *

from . import has_permission
from .schema import AdminSchema, PatchAdminSchema, SuccessfulSchema


class AdminUserApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin})]

    @app.output(AdminSchema)  # type: ignore
    def get(self, uuid):
        """Admin: Get an admin"""
        admin = AdminUser.by_uuid(uuid) or abort(404, "Admin not found")
        if not has_permission(admin):
            abort(403, "you don't have permission to access this admin")
        return admin.to_schema()  # type: ignore

    @app.input(PatchAdminSchema, arg_name='data')  # type: ignore
    @app.output(AdminSchema)  # type: ignore
    def patch(self, uuid, data):
        """Admin: Update an admin"""
        admin = AdminUser.by_uuid(uuid) or abort(404, "Admin not found")
        if not has_permission(admin):
            abort(403, "You don't have permission to access this admin")

        for field in AdminUser.__table__.columns.keys():  # type: ignore
            if field in ['id', 'parent_admin_id']:
                continue
            if field not in data:
                data[field] = getattr(admin, field)
        data['old_uuid'] = uuid
        admin = AdminUser.add_or_update(True, **data) or abort(502, "Unknown issue: Admin is not patched")
        # the add_or_update doesn't update the uuid of AdminUser, so for now just delete old admin after adding new

        return admins

    @app.output(SuccessfulSchema)  # type: ignore
    def delete(self, uuid):
        """Admin: Delete an admin"""
        admin = AdminUser.by_uuid(uuid) or abort(404, "Admin not found")
        if not has_permission(admin):
            abort(403, "You don't have permission to access this admin")
        admin.remove()  # type: ignore
        return {'status': 200, 'msg': 'ok'}




from typing import TYPE_CHECKING
from hiddifypanel.models import hconfig, ConfigEnum
if TYPE_CHECKING:
    import cloudflare

__cf: "cloudflare.Cloudflare"=None   # type: ignore


def __prepare_cf_api_client() -> bool:
    import cloudflare

    '''Prepares cloudflare client if it's not already'''
    global __cf
    if __cf and isinstance(__cf, cloudflare.Cloudflare):
        return True

    if hconfig(ConfigEnum.cloudflare):
        __cf = cloudflare.Cloudflare(api_token=hconfig(ConfigEnum.cloudflare))
        if __cf and isinstance(__cf, cloudflare.Cloudflare):
            return True
    return False


def add_or_update_dns_record(domain: str, ip: str, dns_type: str = "A", proxied: bool = True) -> bool:
    '''This function cloud throw an exception'''
    if not __prepare_cf_api_client():
        return False

    zone_name = __extract_root_domain(domain)
    zone = __get_zone(zone_name)
    if zone:
        record = __get_dns_record(zone, domain,dns_type)
        dns_name = domain[:-len(zone.name)].replace('.', '')
        # if the input domain is root itself
        dns_name = '@' if not dns_name else dns_name
        
        if not record:
            api_res = __cf.dns.records.create(zone_id=zone.id, name=dns_name,type=dns_type,content=ip,proxied=proxied)
        else:
            api_res = __cf.dns.records.edit(zone_id=zone.id,dns_record_id= record.id, name=dns_name,type=dns_type,content=ip,proxied=proxied)

        # validate api response
        if api_res.name == domain and api_res.type == dns_type and api_res.content == ip:
            return True
    return False


def delete_dns_record(domain: str) -> bool:
    '''Deletes a DNS record from cloudflare panel of user'''
    if not __prepare_cf_api_client():
        return False

    zone_name = __extract_root_domain(domain)
    zone = __get_zone(zone_name)
    records = [__get_dns_record(zone, domain,"A"),__get_dns_record(zone, domain,"AAAA")]
    res=False
    
    for record in records:
        if record and zone:
            api_res = __cf.dns.records.delete(dns_record_id=record.id,zone_id=zone.id)
            if api_res.id == record.id:
                res=True
            
    return res


def __get_zone(zone_name: str):
    zones = __cf.zones.list()
    for z in zones:
        if z.name == zone_name:
            return z
    return None


def __get_dns_record(zone, domain: str,dns_type) :
    '''Returns dns record if exists'''
    dns_records = __cf.dns.records.list(zone_id=zone.id)#__cf.zones.get(zone['id']).dns_records(zone['id'])
    for r in dns_records:
        if r.name == domain and r.type==dns_type:
            return r
    return None


def __extract_root_domain(domain: str) -> str:
    domain_parts = domain.split(".")
    if len(domain_parts) > 1:
        return ".".join(domain_parts[-2:])
    else:
        return domain



import ipaddress
from hiddifypanel.auth import login_required, current_account

from hiddifypanel.hutils.flask import hurl_for
from hiddifypanel.models import *
import re
from flask import g  # type: ignore
from markupsafe import Markup

from flask_babel import gettext as __
from flask_babel import lazy_gettext as _
from hiddifypanel.panel.run_commander import Command, commander
from wtforms.validators import Regexp, ValidationError

from hiddifypanel.models import *
from hiddifypanel.panel import hiddify, custom_widgets
from .adminlte import AdminLTEModelView
from hiddifypanel import hutils

from loguru import logger
from flask import current_app
# Define a custom field type for the related domains


# class ConfigDomainsField(SelectField):
#     def __init__(self, label=None, validators=None,*args, **kwargs):
#         kwargs.pop("allow_blank")
#         super().__init__(label, validators,*args, **kwargs)
#         self.choices=[(d.id,d.domain) for d in Doamin.query.filter(Domain.sub_link_only!=True).all()]


class DomainAdmin(AdminLTEModelView):
    # edit_modal = False
    # create_modal = False
    column_hide_backrefs = False

    list_template = 'model/domain_list.html'
    # edit_modal = True
    form_overrides = {'mode': custom_widgets.EnumSelectField}
    form_widget_args = {
        'description': {
            'rows': 100,
            'style': 'font-family: monospace; direction:ltr'
        }
    }
    column_descriptions = dict(
        domain=_("domain.description"),
        mode=_("Direct mode means you want to use your server directly (for usual use), CDN means that you use your server on behind of a CDN provider."),
        cdn_ip=_("config.cdn_forced_host.description"),
        show_domains=_('domain.show_domains_description'),
        alias=_('The name shown in the configs for this domain.'),
        servernames=_('config.reality_server_names.description'),
        sub_link_only=_('This can be used for giving your users a permanent non blockable links.'),
        grpc=_('grpc-proxy.description'),
        download_domain=_('download_domain.description'),
        resolve_ip=_("domain.resolveip.description")
    )
    # create_modal = True
    can_export = False
    form_widget_args = {'show_domains': {'class': 'form-control ltr'},'download_domain': {'class': 'form-control ltr'}}

    form_args = {
        'mode': {'enum': DomainType},
        'show_domains': {
            'query_factory': lambda: Domain.query.filter(     Domain.sub_link_only == False),
        },
        'domain': {
            'validators': [
                Regexp(r'^(\*\.)?([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$|^$|^(\d{1,3}\.){3}\d{1,3}$|^([0-9a-fA-F]{1,4}:){1,7}(:|[0-9a-fA-F]{1,4})$',message=__("Should be a valid domain"))]},
        "cdn_ip": {
            'validators': [
                Regexp(r"(((((25[0-5]|(2[0-4]|1\d|[1-9]|)\d).){3}(25[0-5]|(2[0-4]|1\d|[1-9]|)\d))|^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,}))[ \t\n,;]*\w{3}[ \t\n,;]*)*",message=__("Invalid IP or domain"))]},
        "servernames": {
            'validators': [
                Regexp(r"^([\w-]+\.)+[\w-]+(,\s*([\w-]+\.)+[\w-]+)*$",re.IGNORECASE,_("Invalid REALITY hostnames"))]}}
    column_list = ["domain", "alias", "mode",  "show_domains"]
    column_editable_list = ["alias"]
    # column_filters=["domain","mode"]
    # form_excluded_columns=['work_with']
    column_searchable_list = ["domain", "mode"]
    column_labels = {
        "domain": _("domain.domain"),
        'sub_link_only': _('Only for sublink?'),
        "mode": _("domain.mode"),
        "cdn_ip": _("config.cdn_forced_host.label"),
        'domain_ip': _('domain.ip'),
        'servernames': _('config.reality_server_names.label'),
        'show_domains': _('Show Domains'),
        'alias': _('Alias'),
        'grpc': _('gRPC'),
        "download_domain":_('download_domain.label'),
        'resolve_ip':_("domain.resolveip.label"),
    }

    form_columns = ['mode', 'domain', 'alias', 'servernames', 'cdn_ip', 'resolve_ip', 'show_domains', 'download_domain',]
    
    def _domain_admin_link(view, context, model, name):
        if hiddify.is_fake_domain(model):
            return Markup(f"<span class='badge'>{model.domain}</span>")
        d = model.domain
        if "*" in d:
            d = d.replace("*", hutils.random.get_random_string(5, 15))
        admin_link = hiddify.get_account_panel_link(g.account, d)
        return Markup(
            f'<div class="btn-group"><a href="{admin_link}" class="btn btn-xs btn-secondary">' + _("admin link") +
            f'</a><a href="{admin_link}" class="btn btn-xs btn-info ltr" target="_blank">{model.domain}</a></div>'+
            f'<a href="{hurl_for('admin.Actions:get_domain_ip',domain=model.domain)}"><i class="fa-solid fa-dharmachakra"></i></a>')

    def _domain_ip(view, context, model, name):
        dips = hutils.network.get_domain_ips_cached(model.domain)
        # The get_domain_ip function uses the socket library, which relies on the system DNS resolver. So it may sometimes use cached data, which is not desirable
        # if not dips:
        #     dip = hutils.network.resolve_domain_with_api(model.domain)
        myips = set(hutils.network.get_ips())
        all_res = ""
        for dip in dips:
            if dip in myips and model.mode in [DomainType.direct, DomainType.sub_link_only]:
                badge_type = ''
            elif dip and dip not in myips and model.mode != DomainType.direct:
                badge_type = 'warning'
            else:
                badge_type = 'danger'
            res = f'<span class="badge badge-{badge_type}">{dip}</span>'
            if model.sub_link_only:
                res += f'<span class="badge badge-success">{_("SubLink")}</span>'
            all_res += res
        return Markup(all_res)

    def _show_domains_formater(view, context, model, name):
        if hiddify.is_fake_domain(model):
            return ""
        if not len(model.show_domains):
            return _("All")
        else:
            return Markup(" ".join([hiddify.get_domain_btn_link(d) for d in model.show_domains]))

    column_formatters = {
        # 'domain_ip': _domain_ip,
        'domain': _domain_admin_link,
        'show_domains': _show_domains_formater
    }

    def search_placeholder(self):
        return f"{_('search')} {_('domain.domain')} {_('domain.mode')}"

    # def on_form_prefill(self, form, id):
        # Get the Domain object being edited
        # domain = self.session.query(Domain).get(id)

        # Pre-select the related domains in the checkbox list
        # form.show_domains = [d.id for d in Domain.query.all()]

    # TODO: refactor this function
    def on_model_change(self, form, model, is_created):
        # Sanitize domain input
        model.domain = (model.domain or '').lower().strip()
        if model.download_domain and model.domain==model.download_domain.domain:
            model.download_domain_id=None
            model.download_domain=None
        # Basic validation
        if model.domain == '' and model.mode != DomainType.fake:
            raise ValidationError(_("domain.empty.allowed_for_fake_only"))

        self._validate_not_used_before(model,is_created)
        ipv4_list = hutils.network.get_ips(4)
        ipv6_list = hutils.network.get_ips(6)
        server_ips = [*ipv4_list, *ipv6_list]

        if not server_ips:
            raise ValidationError(_("Couldn't find your ip addresses"))

        # Validate domain based on mode
        if "*" in model.domain and model.mode not in [DomainType.cdn, DomainType.auto_cdn_ip]:
            raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

        cloudflare_updated=self._update_cloudflare(model, ipv4_list,ipv6_list)
        
        if not cloudflare_updated:
            self._validate_domain_ips(model, server_ips)

        # Handle CDN IP settings
        if  model.mode == DomainType.direct and model.cdn_ip:
            model.cdn_ip = ""
            raise ValidationError(_("Specifying CDN IP is only valid for CDN mode"))
            
        if model.mode == DomainType.fake and not model.cdn_ip:
            model.cdn_ip = str(server_ips[0])
            
        if model.cdn_ip:
            try:
                hutils.network.auto_ip_selector.get_clean_ip(str(model.cdn_ip))
            except Exception:
                raise ValidationError(_("Error in auto cdn format"))
                    
        # Update show domains
        if len(model.show_domains) == Domain.query.count():
            model.show_domains = []
                
        # Handle mode-specific settings
        if model.mode == DomainType.old_xtls_direct and not hconfig(ConfigEnum.xtls_enable):
            set_hconfig(ConfigEnum.xtls_enable, True)
            hutils.proxy.get_proxies().invalidate_all()
        elif "reality" in  model.mode:
            self._validate_reality_settings(model, server_ips)
                
            # Signal config update if needed
        old_db_domain = Domain.by_domain(model.domain)
        if is_created or not old_db_domain or old_db_domain.mode != model.mode:
            # return hiddify.reinstall_action(complete_install=False, domain_changed=True)
            hutils.flask.flash_config_success(restart_mode=ApplyMode.apply_config, domain_changed=True)

            

    def _update_cloudflare(self, model, ipv4_list,ipv6_list):
        if hconfig(ConfigEnum.cloudflare) and model.mode not in [DomainType.fake, DomainType.relay, DomainType.reality]:
            try:
                proxied = model.mode in [DomainType.cdn, DomainType.auto_cdn_ip]
                if ipv4_list:
                    hutils.network.cf_api.add_or_update_dns_record(model.domain, str(ipv4_list[0]), "A", proxied=proxied)
                if ipv6_list:
                    hutils.network.cf_api.add_or_update_dns_record(model.domain, str(ipv6_list[0]), "AAAA", proxied=proxied)
                return True
            except Exception as e:
                raise ValidationError(__("cloudflare.error") + f' {e}')
        return False

    def _validate_reality_settings(self, model, server_ips):
        """Validate REALITY protocol settings with proper error handling"""
        if not hconfig(ConfigEnum.reality_enable):
            set_hconfig(ConfigEnum.reality_enable, True)
            hutils.proxy.get_proxies().invalidate_all()

        model.servernames = (model.servernames or model.domain).lower().strip()
        domains_to_check = set()
        for v in [model.domain, model.servernames]:
            domains_to_check.update(d.strip() for d in v.split(",") if d.strip())

        for d in domains_to_check:
            # Check REALITY compatibility
            if not hutils.network.is_domain_reality_friendly(d):
                raise ValidationError(_("Domain is not REALITY friendly!") + f' {d}')

            try:
                if not hutils.network.is_in_same_asn(d, server_ips[0]):
                    domain_ips = hutils.network.get_domain_ips(d)
                    if domain_ips:
                        dip = next(iter(domain_ips))
                        server_asn = hutils.network.get_ip_asn(server_ips[0])
                        domain_asn = hutils.network.get_ip_asn(dip)
                        msg = _("domain.reality.asn_issue")
                        if server_asn or domain_asn:
                            msg += f"<br> Server ASN={server_asn}<br>{d}_ASN={domain_asn}"
                        hutils.flask.flash(msg, 'warning')
            except Exception as e:
                logger.warning(f"ASN check failed for domain {d}: {str(e)}")

        # Check fallback compatibility
        for d in model.servernames.split(","):
            if d.strip() and not hutils.network.fallback_domain_compatible_with_servernames(model.domain, d):
                msg = _("REALITY Fallback domain is not compatible with server names!") + f' {d} != {model.domain}'
                hutils.flask.flash(msg, 'warning')


    def _validate_not_used_before(self, model,is_created):
        configs = get_hconfigs()
        for c in configs:
            if "domain" in c and c not in [ConfigEnum.decoy_domain, ConfigEnum.reality_fallback_domain] and c.category != 'hidden':
                if model.domain == configs[c]:
                    raise ValidationError(_("You have used this domain in: ") + _(f"config.{c}.label"))

        for td in Domain.query.filter(Domain.mode.in_([DomainType.reality,DomainType.special_reality_xhttp,DomainType.special_reality_grpc,DomainType.special_reality_tcp]), Domain.domain != model.domain).all():
            # print(td)
            if td.servernames and (model.domain in td.servernames.split(",")):
                raise ValidationError(_("You have used this domain in: ") + _(f"config.reality_server_names.label") + td.domain)

        if is_created and Domain.query.filter(Domain.domain == model.domain, Domain.child_id == model.child_id).count() > 1:
            raise ValidationError(_("You have used this domain in: "))

    def _validate_domain_ips(self, model, server_ips):
        """Validate domain IP resolution and matching"""
        
        # Skip validation for wildcard or empty domains
        if (model.domain.startswith('*') or not model.domain) and model.mode not in [DomainType.direct]:
            return True
        if model.mode in [DomainType.fake, DomainType.reality, DomainType.relay, DomainType.dnstt]:
            return True
        if "special" in model.mode:
            return True
        # Resolve domain IPs with timeout
        try:
            dips = hutils.network.get_domain_ips(model.domain)
        except Exception as e:
            logger.error(f"Error resolving domain {model.domain}: {str(e)}")
            raise ValidationError(_("Domain cannot be resolved! Please check DNS settings"))
        
        # Validate resolution success
        if not dips:
            raise ValidationError(_("Domain cannot be resolved! Please check DNS settings"))
        
        # Check IP matching based on mode
        domain_ip_matches_server = any(ip in dips for ip in server_ips)
        server_ips_str = ', '.join(map(str, server_ips))
        dips_str = ', '.join(map(str, dips))
    
        if not domain_ip_matches_server and model.mode in [DomainType.direct]:
            raise ValidationError(
                __("Domain IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode",
                    server_ip=server_ips_str, domain_ip=dips_str))
                
        if domain_ip_matches_server and model.mode in [DomainType.cdn, DomainType.relay, DomainType.fake, DomainType.auto_cdn_ip]:
            raise ValidationError(
                __("In CDN mode, Domain IP=%(domain_ip)s should be different to your ip=%(server_ip)s",
                    server_ip=server_ips_str, domain_ip=dips_str))
                
        return True
    
        
    # def after_model_change(self,form, model, is_created):
    #     if model.show_domains.count==0:
    #         db.session.bulk_save_objects(ShowDomain(model.id,model.id))

    def on_model_delete(self, model):
        if len(Domain.query.all()) <= 1:
            raise ValidationError(f"at least one domain should exist")
        if hconfig(ConfigEnum.cloudflare) and model.mode not in [DomainType.fake, DomainType.reality, DomainType.relay] and "special" not in model.mode:
            if not hutils.network.cf_api.delete_dns_record(model.domain):
                hutils.flask.flash(_('cf-delete.failed'), 'warning')  # type: ignore
        model.showed_by_domains = []
        # db.session.commit()
        hutils.flask.flash_config_success(restart_mode=ApplyMode.apply_config, domain_changed=True)

    def after_model_delete(self, model):
        if hutils.node.is_child():
            hutils.node.run_node_op_in_bg(hutils.node.child.sync_with_parent, *[hutils.node.child.SyncFields.domains])

    def after_model_change(self, form, model, is_created):
        if hconfig(ConfigEnum.first_setup):
            set_hconfig(ConfigEnum.first_setup, False)
        if model.need_valid_ssl and "*" not in model.domain:
            commander(Command.get_cert, domain=model.domain)
        if hutils.node.is_child():
            hutils.node.run_node_op_in_bg(hutils.node.child.sync_with_parent, *[hutils.node.child.SyncFields.domains])

    def is_accessible(self):
        if login_required(roles={Role.super_admin, Role.admin})(lambda: True)() != True:
            return False
        return True

    # def form_choices(self, field, *args, **kwargs):
    #     if field.type == "Enum":
    #         return [(enum_value.name, _(enum_value.name)) for enum_value in field.type.__members__.values()]
    #     return super().form_choices(field, *args, **kwargs)

    # @property
    # def server_ips(self):
    #     return hiddify.get_ip(4)

    def get_query(self):
        query = super().get_query()
        return query.filter(Domain.child_id == Child.current().id)



from typing import List
from flask import g, request
from apiflask import abort
from flask_restful import Resource
# from flask_simplelogin import login_required
import datetime

from hiddifypanel.auth import login_required
from hiddifypanel import hutils
from hiddifypanel.models import *
from .tgbot import bot


class SendMsgResource(Resource):
    @login_required({Role.super_admin, Role.admin, Role.agent})
    def post(self, admin_uuid=None):

        if not hconfig(ConfigEnum.telegram_bot_token) or not bot:
            abort(400, 'invalid request')

        msg = request.json
        if not msg or not msg.get('id') or not msg.get('text'):
            abort(400, 'invalid request')

        users = self.get_users_by_identifier(msg['id'])

        res = {}
        for user in users:
            try:
                from hiddifypanel.panel.commercial.telegrambot import Usage
                keyboard = Usage.user_keyboard(user.uuid)
                txt = msg['text'] + "\n\n" + Usage.get_usage_msg(user.uuid)
                print('sending to ', user)
                bot.send_message(user.telegram_id, txt, reply_markup=keyboard)
            except Exception as e:
                res[user.uuid] = {'name': user.name, 'error': f'{e}'}
        if len(res) == 0:
            return {'msg': "success"}
        else:
            return {'msg': 'error', 'res': res}

    def get_users_by_identifier(self, identifier: str | list) -> List[User]:
        '''Returns all users that match the identifier for sending a message to them'''
        # when we are here we must have g.account but ...
        if not hasattr(g, 'account'):
            return []
        
        query = User.query.filter(User.added_by.in_(g.account.recursive_sub_admins_ids()))
        query = query.filter(User.telegram_id is not None, User.telegram_id != 0)

        # user selected many ids as users identifier
        if isinstance(identifier, list):
            return query.filter(User.id.in_(identifier)).all()

        if hutils.convert.is_int(identifier):  # type: ignore
            return [query.filter(User.id == int(identifier)).first() or abort(404, 'The user not found')]  # type: ignore
        if identifier == 'all':
            return query.all()
        if identifier == 'expired':
            return [u for u in query.all() if not u.is_active]
        if identifier == 'active':
            return [u for u in query.all() if u.is_active]
        if identifier == 'offline 1h':
            h1 = datetime.datetime.now() - datetime.timedelta(hours=1)
            return [u for u in query.all() if u.is_active and u.last_online < h1]
        if identifier == 'offline 1d':
            d1 = datetime.datetime.now() - datetime.timedelta(hours=24)
            return [u for u in query.all() if u.is_active and u.last_online < d1]
        if identifier == 'offline 1w':
            d7 = datetime.datetime.now() - datetime.timedelta(days=7)
            return [u for u in query.all() if u.is_active and u.last_online < d7]
        return []



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: hiddifypanel/async/node/test.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

import hiddifypanel.hasync.node.test_pb2


class HelloBase(abc.ABC):

    @abc.abstractmethod
    async def SayHello(self, stream: 'grpclib.server.Stream[hiddifypanel.hasync.node.test_pb2.HelloRequest, hiddifypanel.hasync.node.test_pb2.HelloResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/Hello/SayHello': grpclib.const.Handler(
                self.SayHello,
                grpclib.const.Cardinality.UNARY_UNARY,
                hiddifypanel.hasync.node.test_pb2.HelloRequest,
                hiddifypanel.hasync.node.test_pb2.HelloResponse,
            ),
        }


class HelloStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.SayHello = grpclib.client.UnaryUnaryMethod(
            channel,
            '/Hello/SayHello',
            hiddifypanel.hasync.node.test_pb2.HelloRequest,
            hiddifypanel.hasync.node.test_pb2.HelloResponse,
        )



import json
import copy
from flask import render_template, g
from hiddifypanel import hutils
from hiddifypanel.models import ProxyTransport, ProxyL3, ProxyProto, Domain, User
from flask_babel import gettext as _
from hiddifypanel.models import hconfig, ConfigEnum
from .xray import is_muxable_agent, OUTBOUND_LEVEL


def configs_as_json(domains: list[Domain], user: User, expire_days: int, remarks: str) -> str:
    '''Returns xray configs as json'''
    all_configs = []

    # region show usage
    if hconfig(ConfigEnum.show_usage_in_sublink) and not g.user_agent.get('is_hiddify'):
        # determine usages
        tag = '⏳ ' if user.is_active else '✖ '
        if user.usage_limit_GB < 1000:
            tag += f'{round(user.current_usage_GB,3)}/{str(user.usage_limit_GB).replace(".0","")}GB'
        elif user.usage_limit_GB < 100000:
            tag += f'{round(user.current_usage_GB/1000,3)}/{str(round(user.usage_limit_GB/1000,1)).replace(".0","")}TB'
        else:
            tag += '#No Usage Limit'
        tag += ' 📅 '
        if expire_days < 1000:
            tag += _(f'%(expire_days)s days', expire_days=expire_days)
        else:
            tag += '#No Time Limit'
        tag = tag.strip()

        # add usage as a config
        all_configs.append(
            null_config(tag)
        )
    # endregion

    if not user.is_active:
        # region show status (active/disable)
        tag = '✖ ' + (hutils.encode.url_encode('بسته شما به پایان رسید')
                      if hconfig(ConfigEnum.lang) == 'fa' else 'Package Ended')
        # add user status
        all_configs.append(
            null_config(tag)
        )
        # endregion
    else:
        # TODO: seperate codes to small functions
        # TODO: check what are unsupported protocols in other apps
        unsupported_protos = {}
        unsupported_transport = {}
        if g.user_agent.get('is_v2rayng'):
            # TODO: ensure which protocols are not supported in v2rayng
            unsupported_protos = {ProxyProto.hysteria, ProxyProto.hysteria2,
                                  ProxyProto.tuic, ProxyProto.ssr, ProxyProto.ssh}
            if not hutils.flask.is_client_version(hutils.flask.ClientVersion.v2ryang, 1, 8, 18):
                unsupported_transport = {ProxyTransport.httpupgrade}
                unsupported_protos.update({ProxyProto.wireguard})

        # multiple outbounds needs multiple whole base config not just one with multiple outbounds (at least for v2rayng)
        # https://github.com/2dust/v2rayNG/pull/2827#issue-2127534078
        outbounds = []
        for proxy in hutils.proxy.get_valid_proxies(domains):
            if proxy['proto'] in unsupported_protos:
                continue
            if proxy['transport'] in unsupported_transport:
                continue
            outbound = to_xray(proxy)
            if len(outbound):
                outbounds.append(outbound)

        base_config = json.loads(render_template(
            'base_xray_config.json.j2', remarks=remarks))
        if len(outbounds) > 1:
            for out in outbounds:
                base = copy.deepcopy(base_config)
                base['remarks'] = out['tag']
                base['outbounds'].insert(0, out)
                # if all_configs:
                #     all_configs.insert(0, copy.deepcopy(base_config))
                # else:
                all_configs.append(base)

        else:  # single outbound
            base_config['outbounds'].insert(0, outbounds[0])
            all_configs = base_config

    if not all_configs:
        return ''

    json_configs = json.dumps(all_configs, indent=2, cls=hutils.proxy.ProxyJsonEncoder)
    return json_configs


def to_xray(proxy: dict) -> dict:
    if proxy['proto'] in {ProxyProto.naive,ProxyProto.mieru}:
        return {}
    outbound = {
        'tag': f'{proxy["extra_info"]} {proxy["name"]}',
        'protocol': str(proxy['proto']),
        'settings': {},
        'streamSettings': {},
        # 'mux': {  # default value
        #     # 'enabled': False,
        #     # 'concurrency': -1
        # }
    }

    outbound['protocol'] = 'shadowsocks' if outbound['protocol'] == 'ss' else outbound['protocol']
    # add multiplex to outbound
    add_multiplex(outbound, proxy)

    # add stream setting to outbound
    add_stream_settings(outbound, proxy)

    # add protocol settings to outbound
    add_proto_settings(outbound, proxy)

    return outbound

# region proto settings


def add_proto_settings(base: dict, proxy: dict):
    if proxy['proto'] == ProxyProto.wireguard:
        add_wireguard_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.ss:
        add_shadowsocks_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.vless:
        add_vless_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.vmess:
        add_vmess_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.trojan:
        proxy['password'] = proxy['uuid']
        add_trojan_settings(base, proxy)


def add_wireguard_settings(base: dict, proxy: dict):

    base['settings']['secretKey'] = proxy['wg_pk']
    base['settings']['reversed'] = [0, 0, 0]
    base['settings']['mtu'] = 1380  # optional
    base['settings']['peers'] = [{
        'endpoint': f'{proxy["server"]}:{int(proxy["port"])}',
        'publicKey': proxy["wg_server_pub"],
        "preSharedKey": proxy['wg_psk']

        # 'allowedIPs':'', 'preSharedKey':'', 'keepAlive':'' # optionals
    }]

    # optionals
    # base['settings']['address'] = [f'{proxy["wg_ipv4"]}/32',f'{proxy["wg_ipv6"]}/128']
    # base['settings']['workers'] = 4
    # base['settings']['domainStrategy'] = 'ForceIP' # default


def add_vless_settings(base: dict, proxy: dict):
    base['settings']['vnext'] = [
        {
            'address': proxy['server'],
            'port': proxy['port'],
            "users": [
                {
                    'id': proxy['uuid'],
                    'encryption': 'none',
                    # 'security': 'auto',
                    'flow': proxy.get('flow',''),
                    'level': OUTBOUND_LEVEL
                }
            ]
        }
    ]


def add_vmess_settings(base: dict, proxy: dict):
    base['settings']['vnext'] = [
        {
            "address": proxy['server'],
            "port": proxy['port'],
            "users": [
                {
                    "id": proxy['uuid'],
                    "security": proxy['cipher'],
                    "level": OUTBOUND_LEVEL
                }
            ]
        }
    ]


def add_trojan_settings(base: dict, proxy: dict):
    base['settings']['servers'] = [
        {
            # 'email': proxy['uuid'], optional
            'address': proxy['server'],
            'port': proxy['port'],
            'password': proxy['password'],
            'level': OUTBOUND_LEVEL
        }
    ]


def add_shadowsocks_settings(base: dict, proxy: dict):
    base['settings']['servers'] = [
        {
            'address': proxy['server'],
            'port': proxy['port'],
            'method': proxy['cipher'],
            'password': proxy['password'],
            'uot': True,
            'level': OUTBOUND_LEVEL
            # 'email': '', optional
        }
    ]

# endregion


# region stream settings

def _add_security(base_dict, proxy, tls_info=None):
    if not tls_info:
        tls_info = proxy

    ss = base_dict
    ss['security'] = 'none'  # default

    # security
    if 'reality' in tls_info['mode']:
        ss['security'] = 'reality'
    elif proxy['l3'] in [ProxyL3.tls, ProxyL3.tls_h2, ProxyL3.tls_h2_h1, ProxyL3.h3_quic, ProxyL3.reality]:
        ss['security'] = 'tls'

    # network and transport settings
    # THE CURRENT CODE WORKS BUT THE CORRECT CONDITINO SHOULD BE THIS:
    # ss['security'] == 'tls' or 'xtls' -----> ss['security'] in ['tls','xtls']
    # TODO: FIX THE CONDITION AND TEST CONFIGS ON THE CLIENT SIDE
    if ss['security'] == 'reality':
        # ss['network'] = proxy['transport']
        add_reality_stream(ss, proxy, tls_info)
    elif ss['security'] in ['tls', "xtls"] and proxy['proto'] != ProxyProto.ss:
        ss['tlsSettings'] = {
            'serverName': tls_info['sni'],
            'allowInsecure': tls_info['allow_insecure'],
            'fingerprint': proxy.get('fingerprint'),
            'alpn': [tls_info['alpn']],
            # 'minVersion': '1.2',
            # 'disableSystemRoot': '',
            # 'enableSessionResumption': '',
            # 'pinnedPeerCertificateChainSha256': '',
            # 'certificates': '',
            # 'maxVersion': '1.3', # Go lang sets
            # 'cipherSuites': '', # Go lang sets
            # 'rejectUnknownSni': '', # default is false
        }
        if proxy.get('ech'):
            ss['tlsSettings']['echConfigList'] = [proxy['ech']]


def add_stream_settings(base: dict, proxy: dict):
    ss = base['streamSettings']

    _add_security(ss, proxy, proxy)
    if proxy['l3'] == ProxyL3.kcp:
        ss['network'] = 'kcp'
        add_kcp_stream(ss, proxy)

    if proxy['l3'] == ProxyL3.h3_quic:
        add_quic_stream(ss, proxy)

    if (proxy['transport'] == 'tcp' and ss['security'] != 'reality') or (ss['security'] == 'none' and proxy['transport'] not in [ProxyTransport.httpupgrade, ProxyTransport.WS] and proxy['proto'] != ProxyProto.ss):
        ss['network'] = proxy['transport']
        add_tcp_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.h2 and ss['security'] == 'none' and ss['security'] != 'reality':
        ss['network'] = proxy['transport']
        add_http_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.grpc:
        ss['network'] = proxy['transport']
        add_grpc_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.httpupgrade:
        ss['network'] = proxy['transport']
        add_httpupgrade_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.xhttp:
        ss['network'] = proxy['transport']
        ss['transport'] = "xhttp"
        add_xhttp_stream(ss, proxy)
    if proxy['transport'] == 'ws':
        ss['network'] = proxy['transport']
        add_ws_stream(ss, proxy)

    if proxy['proto'] == ProxyProto.ss:
        ss['network'] = 'tcp'

    # tls fragmentaion
    add_tls_fragmentation_stream_settings(base, proxy)


def add_tcp_stream(ss: dict, proxy: dict):
    
    if proxy.get('params',{}).get('headers',{}).get("type",'')=='none' or proxy['l3'] != ProxyL3.http:
        ss['tcpSettings'] = {
            'header':{'type':'none'}
        }
    else:    
        ss['tcpSettings'] = {
            'header': {
                'type': 'http',
                'request': {
                    'path': [proxy['path']],
                    'method': 'GET',
                    "headers": {
                        "Host": [proxy.get('host')],
                        "User-Agent": [proxy.get('params',{}).get('headers',{}).get('User-Agent')],
                        "Accept-Encoding": ["gzip, deflate"],
                        "Connection": ["keep-alive"],
                        "Pragma": proxy.get('params',{}).get('headers',{}).get('Pragma')
                    },

                }
            }
        }
    # ss['tcpSettings']['header']['request']['headers']



def add_http_stream(ss: dict, proxy: dict):
    ss['httpSettings'] = {
        'host': proxy['host'],
        'path': proxy['path'],
        # 'read_idle_timeout': 10,  # default disabled
        # 'health_check_timeout': 15,  # default is 15
        # 'method': 'PUT',  # default is 15
        # 'headers': {

        # }
    }


def add_ws_stream(ss: dict, proxy: dict):
    ss['wsSettings'] = {
        'path': proxy['path'],
        'headers': {
            "Host": proxy['host']
        }
        # 'acceptProxyProtocol': False,
    }


def add_grpc_stream(ss: dict, proxy: dict):
    ss['grpcSettings'] = {
        # proxy['path'] is equal toproxy['grpc_service_name']
        'serviceName': proxy['path'],
        # by default, the health check is not enabled. may solve some "connection drop" issues
        'idle_timeout': 115,
        'health_check_timeout': 20,  # default is 20
        # 'initial_windows_size': 0,  # 0 means disabled. greater than 65535 means Dynamic Window mechanism will be disabled
        # 'permit_without_stream': False, # health check performed when there are no sub-connections
        # 'multiMode': false, # experimental
    }


def add_httpupgrade_stream(ss: dict, proxy: dict):
    ss['httpupgradeSettings'] = {
        'path': proxy['path'],
        'host': proxy['host'],
        # 'acceptProxyProtocol': '', for inbounds only
    }


def add_xhttp_stream(ss: dict, proxy: dict):
    if ss['transport'] == "xhttp" and g.user_agent.get(hutils.flask.ClientVersion.hiddify_next) and not hutils.flask.is_client_version(hutils.flask.ClientVersion.hiddify_next, 3, 0, 0):
        ss['transport'] = "splithttp"
        ss['splithttpSettings'] = {
            'path': proxy['path'],
            'host': proxy['host'],
            "headers": proxy['params'].get('headers', {})
        }
    else:
        _add_xhttp_details(ss, proxy)

        


def _add_xhttp_details(ss: dict, proxy: dict):
    ss['network'] = "xhttp"
    ss['xhttpSettings'] = {
        'path': proxy['path'],
        'host': proxy['host'],
        'mode':proxy['xhttp_mode'],
        "extra": {
            "headers": proxy['params'].get('headers', {})
        }
    }
    if proxy.get("download"):
        
        dlsettings = {
            "address":proxy['download'].get("server"),
            "port":proxy['port']
        }
        _add_xhttp_details(dlsettings, proxy['download'])
        _add_security(dlsettings, proxy, proxy['download'])
        
        ss['xhttpSettings']['extra']['downloadSettings']=dlsettings


def add_kcp_stream(ss: dict, proxy: dict):
    # TODO: fix server side configs first
    ss['kcpSettings'] = {}
    return
    ss['kcpSettings'] = {
        'seed': proxy['path'],
        # 'mtu': 1350, # optional, default value is written
        # 'tti': 50, # optional, default value is written
        # 'uplinkCapacity': 5, # optional, default value is written
        # 'downlinkCapacity': 20, # optional, default value is written
        # 'congestion':False, # optional, default value is written
        # 'readBufferSize': 2,# optional, default value is written
        # 'writeBufferSize':2 # optional, default value is written
        # 'header': { # must be same as server (hiddify doesn't use yet)
        #     'type': 'none'  # choices: none(default), srtp, utp, wechat-video, dtls, wireguards
        # }
    }


def add_quic_stream(ss: dict, proxy: dict):
    # TODO: fix server side configs first
    return

    ss['quicSettings'] = {
        'security': 'chacha20-poly1305',
        'key': proxy['path'],
        'header': {
            'type': 'none'
        }
    }


def add_reality_stream(ss: dict, proxy: dict, domain_info: dict):
    ss['realitySettings'] = {
        'serverName': domain_info['sni'],
        'fingerprint': proxy['fingerprint'],
        'shortId': domain_info['reality_short_id'],
        'publicKey': domain_info['reality_pbk'],
        'show': False,
    }


def add_tls_fragmentation_stream_settings(base: dict, proxy: dict):
    '''Adds tls fragment in the outbounds if tls fragmentation is enabled'''
    if base['streamSettings']['security'] in ['tls', 'reality']:
        if proxy.get('tls_fragment_enable'):
            base['streamSettings']['sockopt'] = {
                'dialerProxy': 'fragment',
                'tcpKeepAliveIdle': 100,
                # recommended to be enabled with "tcpMptcp": true.
                'tcpNoDelay': True,
                "mark": 255
                # 'tcpFastOpen': True, # the system default setting be used.
                # 'tcpKeepAliveInterval': 0, # 0 means default GO lang settings, -1 means not enable
                # 'tcpcongestion': bbr, # Not configuring means using the system default value
                # 'tcpMptcp': True, # need to be enabled in both server and client configuration (not supported by panel yet)
            }

# endregion


def add_multiplex(base: dict, proxy: dict):
    if proxy.get('mux_enable') != "xray" or not is_muxable_agent(proxy):
        return

    concurrency = proxy['mux_max_connections']
    if concurrency and concurrency > 0:
        base['mux'] = {'enabled': True,
                       'concurrency': concurrency,
                       'xudpConcurrency': concurrency,
                       'xudpProxyUDP443': 'reject',
                       }


def null_config(tag: str) -> dict:
    base_config = json.loads(render_template(
        'base_xray_config.json.j2', remarks=tag))
    base_config['outbounds'][0]["protocol"] = "blackhole"
    return base_config




from strenum import StrEnum
from enum import auto
from sqlalchemy import Column, String, Integer, Boolean, Enum, ForeignKey

from hiddifypanel.database import db

from sqlalchemy.types import JSON


class ProxyTransport(StrEnum):
    h2 = auto()
    grpc = auto()
    # XTLS = auto()
    faketls = auto()
    shadowtls = auto()
    restls1_2 = auto()
    restls1_3 = auto()
    # h1=auto()
    WS = auto()
    tcp = auto()
    ssh = auto()
    httpupgrade = auto()
    xhttp = auto()
    custom = auto()
    shadowsocks = auto()
    udp = auto()


class ProxyCDN(StrEnum):
    CDN = auto()
    direct = auto()
    Fake = auto()
    relay = auto()


class ProxyProto(StrEnum):
    vless = auto()
    trojan = auto()
    vmess = auto()
    ss = auto()
    v2ray = auto()
    ssr = auto()
    ssh = auto()
    tuic = auto()
    hysteria = auto()
    hysteria2 = auto()
    wireguard = auto()
    naive = auto()
    mieru = auto()
    dnstt = auto()


class ProxyL3(StrEnum):
    tls = auto()
    tls_h2 = auto()
    tls_h2_h1 = auto()
    h3_quic = auto()
    reality = auto()
    http = auto()
    kcp = auto()
    ssh = auto()
    udp = auto()
    custom = auto()


class Proxy(db.Model):  # type: ignore
    id = Column(Integer, primary_key=True, autoincrement=True)
    child_id = Column(Integer, ForeignKey('child.id'), default=0)
    name = Column(String(200), nullable=False, unique=False)
    enable = Column(Boolean, nullable=False)
    proto = Column(Enum(ProxyProto), nullable=False)
    l3 = Column(Enum(ProxyL3), nullable=False)
    transport = Column(Enum(ProxyTransport), nullable=False)
    cdn = Column(Enum(ProxyCDN), nullable=False)
    params = Column(JSON,default={})

    @property
    def enabled(self):
        return self.enable * 1

    def to_dict(self):
        return {
            'name': self.name,
            'enable': self.enable,
            'proto': self.proto,
            'l3': self.l3,
            'transport': self.transport,
            'cdn': self.cdn,
            'child_unique_id': self.child.unique_id if self.child else '',
            'params': self.params
        }

    def __str__(self):
        return str(self.to_dict())

    @staticmethod
    def add_or_update(commit=True, child_id=0, **proxy):
        dbproxy = Proxy.query.filter(Proxy.name == proxy['name']).first()
        if not dbproxy:
            dbproxy = Proxy()
            db.session.add(dbproxy)  # type: ignore
        dbproxy.enable = proxy['enable']
        dbproxy.name = proxy['name']
        dbproxy.proto = proxy['proto']
        if proxy['transport']=="splithttp":
            proxy['transport']="xhttp"
        dbproxy.transport = proxy['transport']
        dbproxy.cdn = proxy['cdn']
        dbproxy.l3 = proxy['l3']
        dbproxy.params=proxy['params']
        dbproxy.child_id = child_id
        if commit:
            db.session.commit()  # type: ignore

    @staticmethod
    def from_schema(schema):
        return schema.dump(Proxy())

    def to_schema(self):
        proxy_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import ProxySchema
        return ProxySchema().load(proxy_dict)

    @staticmethod
    def bulk_register(proxies, commit=True, force_child_unique_id: str | None = None):
        from hiddifypanel.panel import hiddify
        for proxy in proxies:
            child_id = hiddify.get_child(unique_id=force_child_unique_id)
            Proxy.add_or_update(commit=False, child_id=child_id, **proxy)
        if commit:
            db.session.commit()  # type: ignore



import json
import os
import re

import requests

from .abstract_driver import DriverABS
from hiddifypanel.models import User, hconfig, ConfigEnum
from hiddifypanel.panel.run_commander import Command, commander
import redis


USERS_USAGE = "tele:users-usage"


class TelemtApi(DriverABS):
    def get_redis_client(self):
        if not hasattr(self, 'redis_client'):
            self.redis_client = redis.from_url(os.environ.get("REDIS_URI_SSH",""))

        return self.redis_client

    def is_enabled(self) -> bool:
        return hconfig(ConfigEnum.telegram_enable) and hconfig(ConfigEnum.telegram_lib)=="telemt"

    def __init__(self) -> None:
        super().__init__()
        self.tg_uuid_map:dict[str,str]={}

    def __load_tg_uuid_map(self):
        from hiddifypanel.database import db
        users = db.session.query(User).all()
        self.tg_uuid_map={u.uuid.replace("-",""): u.uuid for u in users}

    def __convert_tg_to_uuid(self,pubkeys):
        res={}
        can_reload_map=True
        for key in pubkeys:
            if uuid:=self.tg_uuid_map.get(key):
                res[key]=uuid
            elif can_reload_map:
                self.__load_tg_uuid_map()
                can_reload_map=False
                if uuid:=self.tg_uuid_map.get(key):
                    res[key]=uuid
        return res
    def get_metric(self):
        resp = requests.get("http://localhost:10087/metrics", timeout=5)
        resp.raise_for_status()
        return resp.text

    def __get_tg_usages(self) -> dict:
        raw_output = self.get_metric()
        data = {}

        # Example lines:
        # telemt_user_octets_from_client{user="uuid"} 1983
        # telemt_user_octets_to_client{user="uuid"} 2171

        pattern = re.compile(
            r'telemt_user_octets_(from_client|to_client)\{user="([^"]+)"\}\s+(\d+)'
        )

        for line in raw_output.splitlines():
            match = pattern.search(line)
            if not match:
                continue

            direction, user, value = match.groups()
            value = int(value)

            if user not in data:
                data[user] = {"down": 0, "up": 0}

            if direction == "from_client":
                data[user]["up"] = value
            else:  # to_client
                data[user]["down"] = value

        return data

    def __get_local_usage(self) -> dict:
        usage_data = self.get_redis_client() .get(USERS_USAGE)
        if usage_data:
            return json.loads(usage_data)

        return {}

    def __sync_local_usages(self) -> dict:
        local_usage = self.__get_local_usage()
        tg_usage = self.__get_tg_usages()
        
        res = {}
        # remove local usage that is removed from wg usage
        for local_uuid in local_usage.copy().keys():
            if local_uuid not in tg_usage:
                del local_usage[local_uuid]

        
        uuid_map = self.__convert_tg_to_uuid(tg_usage.keys())
        for tg_uuid, tg_usage in tg_usage.items():
            uuid = uuid_map.get(tg_uuid)
            
            if not local_usage.get(tg_uuid):
                local_usage[tg_uuid] = {"uuid": uuid, "usage": tg_usage}
                continue
            res[uuid] = self.calculate_reset(local_usage[tg_uuid]['usage'], tg_usage)
            local_usage[tg_uuid] = {"uuid": uuid, "usage": tg_usage}

        self.get_redis_client().set(USERS_USAGE, json.dumps(local_usage))

        return res

    def calculate_reset(self, last_usage: dict, current_usage: dict) -> dict:
        res = {
            'up': current_usage['up'] - last_usage['up'],
            'down': current_usage['down'] - last_usage['down'],
        }

        if res['up'] < 0:
            res['up'] = 0
        if res['down'] < 0:
            res['down'] = 0
        return res

    def get_enabled_users(self):
        if not self.is_enabled():
            return {}
        usages = self.__get_tg_usages()
        new_wg_pubs = set(usages.keys())
        old_usages = self.__get_local_usage()
        old_wg_pubs = set(old_usages.keys())
        enabled = {u['uuid']: 1 for u in old_usages.values()}
        not_included = new_wg_pubs - old_wg_pubs
        if not_included:
            users = User.query.filter(User.wg_pub.in_(not_included).all())
            for u in users:
                enabled[u.uuid] = 1

        return enabled

    def add_client(self, user):
        pass

    def remove_client(self, user):
        pass

    def get_all_usage(self, reset=True):
        if not self.is_enabled():
            return {}
        all_usages = self.__sync_local_usages()
        res = {}
        for uuid,use in all_usages.items():
            # if use := all_usages.get(u.wg_pub):
                res[uuid] = use['up'] + use['down']
            # else:
            #     res[u] = 0
        return res



import datetime
import json
from flask import request, g
from hiddifypanel import hutils
from hiddifypanel.models import ProxyTransport, ProxyL3, ProxyProto, Domain, User, ConfigEnum, hconfig
from flask_babel import gettext as _
from urllib.parse import urlencode, quote
OUTBOUND_LEVEL = 8


def is_muxable_agent(proxy: dict) -> bool:
    if not proxy.get('mux_enable'):
        return False
    if proxy.get('mux_enable') == "xray" and g.user_agent.get('is_singbox'):
        return False
    if proxy.get('mux_enable') == "singbox" and not g.user_agent.get('is_singbox'):
        return False
    return True


def to_link(proxy: dict) -> str | dict:
    if 'error' in proxy:
        return proxy

    orig_name_link = (proxy['extra_info'] + " " + proxy["name"]).strip()
    name_link = hutils.encode.url_encode(orig_name_link)
    if proxy['proto']=="dnstt":
        resolvers="&".join([f"resolver={r}" for r in proxy['resolvers']])
        dnstt=f'dnstt://?tunnel_per_resolver={proxy["tunnel_per_resolver"]}&{resolvers}&domain={proxy["sni"]}&publicKey={proxy["public_key"]}'
        return f'socks://{proxy["uuid"]}:{proxy["password"]}@localhost:0#{name_link} -> {dnstt}'
    if proxy['proto'] == "naive":
        naive = f'naive://{proxy["uuid"]}:{proxy["password"]}@{proxy["server"]}:{proxy["port"]}/?security=tls&sni={proxy["sni"]}&uot=1&header=hiddify-naive-secret:{proxy["path"]}'
        if proxy.get('quic'):
            naive += "&quic=1"
        return f'{naive}#{name_link}'

    if proxy['proto'] == "mieru":
        mieru = f'mieru://{proxy["uuid"]}:{proxy["password"]}@{proxy["server"]}/?handshake-mode={proxy["handshake"]}&mtu=1400&multiplexing={proxy["multiplexing"]}'
        for port in proxy["tcp_ports"]:
            if port:
                mieru += f"&port={port}&protocol=TCP"
        for port in proxy["udp_ports"]:
            if port:
                mieru += f"&port={port}&protocol=UDP"

        return f'{mieru}#{name_link}'

    if proxy['proto'] == 'vmess':
        # print(proxy)
        vmess_type = None
        if proxy["transport"] == 'tcp':
            vmess_type = 'http'
        if 'grpc_mode' in proxy:
            vmess_type = proxy['grpc_mode']
        vmess_data = {"v": "2",
                      "ps": orig_name_link,
                      "add": proxy['server'],
                      "port": proxy['port'],
                      "id": proxy["uuid"],
                      "aid": 0,
                      "scy": proxy['cipher'],
                      "net": proxy["transport"],
                      "type": vmess_type or "none",
                      "host": proxy.get("host", ""),
                      "alpn": proxy.get("alpn", "h2,http/1.1"),
                      "path": proxy["path"] if "path" in proxy else "",
                      "tls": "tls" if "tls" in proxy["l3"] or "quic" in proxy['l3'] else "",
                      "sni": proxy["sni"],
                      "fp": proxy["fingerprint"]
                      }
        if proxy.get('ech'):
            vmess_data['ech'] = proxy['ech']
        if 'reality' in proxy["l3"]:
            vmess_data['tls'] = "reality"
            vmess_data['pbk'] = proxy['reality_pbk']
            vmess_data['sid'] = proxy['reality_short_id']
        if proxy.get('transport') in {ProxyTransport.xhttp}:
            vmess_data['core'] = 'xray'
            _add_xhttp_extra(vmess_data, proxy)
        if vmess_data.get('tls') == 'tls' and proxy.get('ech'):
            vmess_data['ech'] = proxy['ech']
        add_tls_tricks_to_dict(vmess_data, proxy)
        add_mux_to_dict(vmess_data, proxy)

        return "vmess://" + hutils.encode.do_base_64(f'{json.dumps(vmess_data, cls=hutils.proxy.ProxyJsonEncoder)}')
    if proxy['proto'] == 'ssh':
        # baseurl = 'ssh://'
        # if g.user_agent.get('is_streisand'):
        #     streisand_ssh = hutils.encode.do_base_64(f'{proxy["uuid"]}:0:{proxy["private_key"]}::@{proxy["server"]}:{proxy["port"]}')
        #     baseurl += f'{streisand_ssh}#{name_link}'
        # else:
        hk = ",".join(proxy["host_keys"])
        pk = proxy["private_key"]
        q = {
            'file': 'ssh',
            'hk': hk,
            'pk': pk,
            'private_key': pk,
            'authentication': 0,
            'passphrase': '',
        }
        return f"ssh://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}/?{urlencode(q, quote_via=quote)}#{name_link}"
        # baseurl += f'{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}/?file=ssh&pk={pk}&hk={hk}&private_key={pk}&authentication=0&passphrase#{name_link}'

        # return baseurl
    if proxy['proto'] == "ssr":
        baseurl = f'ssr://{proxy["cipher"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}'
        return baseurl
    if proxy['proto'] in ['ss', 'v2ray']:
        baseurl = f'ss://{hutils.encode.do_base_64(proxy["cipher"] + ":" + proxy["password"])}@{proxy["server"]}:{proxy["port"]}'

        if proxy['transport'] == 'shadowsocks':
            return f'{baseurl}#{name_link}'
        if proxy['transport'] == 'faketls':
            return f'{baseurl}?plugin=obfs-local&obfs-host={proxy["fakedomain"]}&obfs=http&udp-over-tcp=true#{name_link}'
        if proxy['transport'] == 'shadowtls':
            return "ShadowTLS is Not Supported for this platform"
            # return f'{baseurl}?plugin=v2ray-plugin&path={proxy["proxy_path"]}&host={proxy["fakedomain"]}&udp-over-tcp=true#{name_link}'
        if proxy['proto'] == 'v2ray':
            return f'{baseurl}?plugin=v2ray-plugin&mode=websocket&path={proxy["proxy_path"]}&host={proxy["sni"]}&tls&udp-over-tcp=true#{name_link}'

    if proxy['proto'] == 'tuic':
        baseurl = f'tuic://{proxy["uuid"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?congestion_control=cubic&udp_relay_mode=native&sni={proxy["sni"]}&alpn=h3'
        if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
            baseurl += "&allow_insecure=1"
        return f"{baseurl}#{name_link}"
    if proxy['proto'] == 'hysteria2':
        baseurl = f'hysteria2://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?hiddify=1&obfs=salamander&obfs-password={proxy["hysteria_obfs_password"]}&sni={proxy["sni"]}'
        if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
            baseurl += "&insecure=1&allow_insecure=1"
        return f"{baseurl}#{name_link}"
    if proxy['proto'] == ProxyProto.wireguard:
        if g.user_agent.get('is_streisand'):
            query = {
                "private_key": proxy["wg_pk"],
                "peer_public_key": proxy["wg_server_pub"],
                "pre_shared_key":   proxy["wg_psk"],
                "reserved": "0,0,0"
            }
            return f'wireguard://{proxy["server"]}:{proxy["port"]}?{urlencode(query)}#{name_link}'
        else:
            # hiddify_format =
            # f'wg://{proxy["server"]}:{proxy["port"]}/?pk={proxy["wg_pk"]}&local_address={proxy["wg_ipv4"]}/32&peer_pk={proxy["wg_server_pub"]}&pre_shared_key={proxy["wg_psk"]}&workers=4&mtu=1380&reserved=0,0,0&ifp={proxy["wg_noise_trick"]}#{name_link}'

            query = {
                "privateKey": proxy["wg_pk"],
                "publicKey": proxy["wg_server_pub"],
                "presharedKey":   proxy["wg_psk"],
                "reserved": "0,0,0",
                "ip":"10.0.0.1",
                "mtu":"1380",
                "keepalive":"30",
                "udp":1,
                "ifp":proxy["wg_noise_trick"]
            }
            return f'wg://{proxy["server"]}:{proxy["port"]}?{urlencode(query)}#{name_link}'

    baseurl = f'{proxy["proto"]}://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}'

    q = {
        'hiddify': 1,
        'sni': proxy['sni'],
        'type': proxy['transport'],
        'alpn': proxy['alpn']
    }

    # the ray2sing supports vless, vmess and trojan tls tricks and mux
    # the vmess handled already

    add_mux_to_dict(q, proxy)
    add_tls_tricks_to_dict(q, proxy)
    if "path" in proxy:
        q['path'] = proxy["path"]
    if "host" in proxy:
        q['host'] = proxy["host"]
    # infos+=f'&alpn={proxy["alpn"]}'

    if "grpc" == proxy["transport"]:
        q['serviceName'] = proxy["grpc_service_name"]
        q['mode'] = proxy["grpc_mode"]
    # print(proxy['cdn'],proxy["transport"])
    if request.args.get("fragment"):
        q['fragment'] = request.args.get("fragment")  # type: ignore
    if "ws" == proxy["transport"] and proxy['cdn'] and request.args.get("fragment_v1"):
        q['fragment_v1'] = request.args.get("fragment_v1")  # type: ignore
    if 'vless' == proxy['proto']:
        q['encryption'] = 'none'

    if proxy.get('fingerprint', 'none') != 'none':
        q['fp'] = proxy['fingerprint']
    if proxy.get('transport') in {ProxyTransport.xhttp}:
        q['core'] = 'xray'
        _add_xhttp_extra(q, proxy)
    if proxy['l3'] != 'quic':
        if proxy.get('params', {}).get('headers', {}).get("type", '') == 'none' or proxy['l3'] != ProxyL3.http:
            q['headerType'] = 'none'
        else:
            # if proxy.get('l3') != ProxyL3.reality and (proxy.get('transport') in {ProxyTransport.tcp, ProxyTransport.httpupgrade, ProxyTransport.xhttp}) and proxy['proto'] in [ProxyProto.vless, ProxyProto.trojan]:
            q['headerType'] = 'http'

    if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
        q['allowInsecure'] = 'true'
        q['insecure'] = 'true'
    if proxy.get('flow'):
        q['flow'] = proxy["flow"]

    if 'reality' in proxy["l3"]:
        q['security'] = 'reality'
        q['pbk'] = proxy['reality_pbk']
        q['sid'] = proxy['reality_short_id']

    elif 'tls' in proxy['l3'] or "quic" in proxy['l3']:
        q['security'] = 'tls'

    elif proxy['l3'] == 'http':
        q['security'] = 'none'
    if proxy.get('ech'):
        q['ech'] = proxy['ech']
    if proxy.get('transport') not in {ProxyTransport.xhttp}:
        for k, v in proxy.get('params', {}).items():
            if k not in q and k != "download":
                q[k] = v
    return f"{baseurl}?{urlencode(q, quote_via=quote)}#{name_link}"


def make_v2ray_configs(domains: list[Domain], user: User, expire_days: int, ip_debug=None) -> str:
    res = []

    if hconfig(ConfigEnum.show_usage_in_sublink) and not g.user_agent.get('is_hiddify'):

        fake_ip_for_sub_link = datetime.datetime.now().strftime(f"%H.%M--%Y.%m.%d.time:%H%M")
        # if ua['app'] == "Fair1":
        #     res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{round(user.current_usage_GB,3)}/{user.usage_limit_GB}GB_Remain:{expire_days}days')
        # else:

        # res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hutils.encode.url_encode(profile_title)}')

        name = '⏳ ' if user.is_active else '✖ '
        if user.usage_limit_GB < 1000:
            name += f'{round(user.current_usage_GB, 3)}/{str(user.usage_limit_GB).replace(".0", "")}GB'
        elif user.usage_limit_GB < 100000:
            name += f'{round(user.current_usage_GB/1000, 3)}/{str(round(user.usage_limit_GB/1000, 1)).replace(".0", "")}TB'
        else:
            res.append("#No Usage Limit")
        name += " 📅 "
        if expire_days < 1000:
            name += _(f'%(expire_days)s days', expire_days=expire_days)
        else:
            res.append("#No Time Limit")

        name = name.strip()
        if len(name) > 3:
            res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hutils.encode.url_encode(name)}')

    if g.user_agent.get('is_browser') and ip_debug:
        res.append(f'#Hiddify auto ip: {ip_debug}')

    if not user.is_active:

        if hconfig(ConfigEnum.lang) == 'fa':
            res.append('trojan://1@1.1.1.1#' + hutils.encode.url_encode('✖ بسته شما به پایان رسید'))
        else:
            res.append('trojan://1@1.1.1.1#' + hutils.encode.url_encode('✖ Package Ended'))
        return "\n".join(res)

    for pinfo in hutils.proxy.get_valid_proxies(domains):
        link = to_link(pinfo)
        if 'msg' not in link:
            res.append(link)
    return "\n".join(res)


def add_tls_tricks_to_dict(d: dict, proxy: dict):
    if proxy.get('tls_fragment_enable'):
        # if g.user_agent.get('is_shadowrocket'):
        #     d['fragment'] = f'1,{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]}'
        # else:

        d['fragment'] = f'{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]},{proxy.get("tls_fragment_packets", "tlshello")}'
        # if g.user_agent.get('is_streisand'):
        # else:
        #     d['fragment'] = f'tlshello,{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]}'

    if proxy.get("tls_mixed_case"):
        d['mc'] = 1
    if proxy.get("tls_padding_enable"):
        d['padsize'] = proxy["tls_padding_length"]


def add_mux_to_dict(d: dict, proxy):
    if not is_muxable_agent(proxy):
        return

    # according to github.com/hiddify/ray2sing/
    d['muxtype'] = proxy["mux_protocol"]
    d['muxmaxc'] = proxy["mux_max_connections"]
    d['mux'] = proxy['mux_min_streams']
    d['muxsmax'] = proxy["mux_max_streams"]
    d['muxpad'] = proxy["mux_padding_enable"]

    if proxy.get('mux_brutal_enable'):
        d['muxup'] = proxy["mux_brutal_up_mbps"]
        d['muxdown'] = proxy["mux_brutal_down_mbps"]


def _add_xhttp_extra(d: dict, proxy):
    from .xrayjson import _add_xhttp_details
    xhttp_dict = {}
    _add_xhttp_details(xhttp_dict, proxy)
    d['extra'] = json.dumps(xhttp_dict['xhttpSettings']['extra'], separators=(',', ':'))



from flask import request, g
import redis
# from hiddifypanel.cache import cache
from hiddifypanel.models import *

import flask_bootstrap
from flask_babel import Babel
from flask_session import Session

import datetime

from dotenv import dotenv_values
import os
import sys
from werkzeug.middleware.proxy_fix import ProxyFix
from loguru import logger
from sonora.wsgi import grpcWSGI



def init_app(app):
        from hiddifypanel import auth
        app.config["PREFERRED_URL_SCHEME"] = "https"
        app.wsgi_app = ProxyFix(
            app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1,
        )
        

        app.secret_key="asdsad"
        app.servers = {
            'name': 'current',
            'url': '',
        }  # type: ignore
        app.info = {
            'description': 'Hiddify is a free and open source software. It is as it is.',
            'termsOfService': 'https://hiddify.com',
            'contact': {
                'name': 'API Support',
                'url': 'https://www.hiddify.com/support',
                'email': 'panel@hiddify.com'
            },
            'license': {
                'name': 'Creative Commons Zero v1.0 Universal',
                'url': 'https://github.com/hiddify/Hiddify-Manager/blob/main/LICENSE'
            }
        }
        # setup flask server-side session
        # app.config['APPLICATION_ROOT'] = './'
        # app.config['SESSION_COOKIE_DOMAIN'] = '/'
        

        app.jinja_env.line_statement_prefix = '%'
        from hiddifypanel import hutils
        app.jinja_env.filters['b64encode'] = hutils.encode.do_base_64
        app.view_functions['admin.static'] = {}  # fix bug in apiflask
        flask_bootstrap.Bootstrap4(app)

        def get_locale():
            # Put your logic here. Application can store locale in
            # user profile, cookie, session, etc.
            if "admin" in request.base_url:
                g.locale = hconfig(ConfigEnum.admin_lang) or 'en'
            else:
                g.locale = auth.current_account.lang or hconfig(ConfigEnum.lang) or 'en'
            return g.locale
        app.jinja_env.globals['get_locale'] = get_locale
        babel = Babel(app, locale_selector=get_locale)
        
        app.config['SESSION_TYPE'] = 'redis'
        
        app.config['SESSION_REDIS'] = redis.from_url(os.environ['REDIS_URI_MAIN'])
        app.config['SESSION_PERMANENT'] = True
        app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(days=10)
        app.security_schemes = {  # equals to use config SECURITY_SCHEMES
            'Hiddify-API-Key': {
                'type': 'apiKey',
                'in': 'header',
                'name': 'Hiddify-API-Key',
            }
        }
        Session(app)
        app.wsgi_app = grpcWSGI(app.wsgi_app)


from typing import Optional
from sqlalchemy.orm import  sessionmaker
from sqlalchemy.orm import as_declarative, declared_attr,relationship
import sqlalchemy.orm as sa_orm

# from sqlalchemy_utils import UUIDType
import re
import os
from sqlalchemy import Row, create_engine, text, Sequence
import sqlalchemy as sa


# class SQLAlchemy:
    
#     def __init__(self):
#         self.engine = create_engine(os.environ.get("SQLALCHEMY_DATABASE_URI"))
#         self.session_maker = sessionmaker(bind=self.engine)
#         self.session=self.session_maker()
#         @as_declarative()
#         class Base:
#             @declared_attr
#             def __tablename__(cls):
#                 return cls.__name__.lower()

#             @classmethod
#             @property
#             def query(cls):
#                 return self.session.query(cls)
            
    
#         self.Query=sa_orm.Query
#         self.Model=Base
#         self.Table=sa.Table
#         self.Column=sa.Column
#         self.Integer=sa.Integer
#         self.ForeignKey=sa.ForeignKey

    # def _set_rel_query(self, kwargs) -> None:
    #         """Apply the extension's :attr:`Query` class as the default for relationships
    #         and backrefs.

    #         :meta private:
    #         """
    #         kwargs.setdefault("query_class", self.Query)

    #         if "backref" in kwargs:
    #             backref = kwargs["backref"]

    #             if isinstance(backref, str):
    #                 backref = (backref, {})

    #             backref[1].setdefault("query_class", self.Query)

        
    # def relationship(
    #         self, *args, **kwargs
    #     ) :
          
    #         self._set_rel_query(kwargs)
    #         return sa_orm.relationship(*args, **kwargs)
from flask_sqlalchemy import SQLAlchemy
    

db = SQLAlchemy()
# db.UUID = UUIDType  # type: ignore

def init_no_flask():
    engine = create_engine(os.environ.get("SQLALCHEMY_DATABASE_URI"))
    db.session = sessionmaker(bind=engine)()

def init_app(app):
    
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    db.init_app(app)
    with app.app_context():
        from hiddifypanel.panel.init_db import init_db
        init_db()
        


def db_execute(query: str, return_val: bool = False, commit: bool = False, **params):
    # print(params)
    q = db.session.execute(text(query), params)
    if commit:
        db.session.commit()
    if return_val:
        return q.fetchall()

    # with db.engine.connect() as connection:
    #     res = connection.execute(text(query), params)
    #     connection.commit()s
    # return res




from flask import g
from flask_babel import lazy_gettext as _
from typing import List
from loguru import logger

from hiddifypanel.models import Child, AdminUser, ConfigEnum, Domain, hconfig, Domain
from hiddifypanel import hutils
from hiddifypanel.panel.commercial.restapi.v2.child.schema import RegisterWithParentInputSchema
from .api_client import NodeApiClient, NodeApiErrorSchema

from hiddifypanel.cache import cache


def request_childs_to_sync():
    for c in Child.query.filter(Child.id != 0).all():
        if not request_child_to_sync(c):
            logger.error(f'{c.name}: {_("parent.sync-req-failed")}')
            hutils.flask.flash(f'{c.name}: ' + _('parent.sync-req-failed'), 'danger') # just for debug


def request_child_to_sync(child: Child) -> bool:
    '''Requests to a child to sync itself with the current panel'''
    child_domain = Domain.get_panel_link(child.id)  # type:ignore
    if not child_domain:
        logger.error(f"Child {child.name} has no valid domain")
        return False

    child_admin_proxy_path = hconfig(ConfigEnum.proxy_path_admin, child.id)
    base_url = f'https://{child_domain}/{child_admin_proxy_path}'
    path = '/api/v2/child/sync-parent/'
    res = NodeApiClient(base_url).post(path, payload=None, output=dict)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while requesting child {child.name} to sync: {res.msg}")
        return False
    if res['msg'] == 'ok':
        logger.success(f"Successfully requested child {child.name} to sync")
        cache.invalidate_all_cached_functions()
        return True

    logger.error(f"Request to child {child.name} to sync failed")
    return False

# before using this function should check child version


# TODO: not used
def request_chlid_to_register(name: str, child_link: str, apikey: str) -> bool:
    '''Requests to a child to register itself with the current panel'''
    if not child_link or not apikey:
        logger.error("Child link or apikey is empty")
        return False
    domain = Domain.get_panel_link()
    if not domain:
        logger.error("Domain is empty")
        return False
    from hiddifypanel.panel import hiddify

    payload = RegisterWithParentInputSchema()
    payload.parent_panel = hiddify.get_account_panel_link(AdminUser.by_uuid(g.account.uuid), domain)  # type: ignore
    payload.apikey = payload.name = hconfig(ConfigEnum.unique_id)

    logger.debug(f"Requesting child {name} to register")
    res = NodeApiClient(child_link, apikey).post('/api/v2/child/register-parent/', payload, dict)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while requesting child {name} to register: {res.msg}")
        return False

    if res['msg'] == 'ok':
        logger.success(f"Successfully requested child {name} to register")
        cache.invalidate_all_cached_functions()
        return True

    logger.error(f"Request to child {name} to register failed")
    return False


def is_child_domain_active(child: Child, domain: Domain) -> bool:
    '''Checks whether a child's domain is responsive'''
    if not domain.need_valid_ssl:
        return False
    child_admin_proxy_path = hconfig(ConfigEnum.proxy_path_admin, child.id)
    if not child_admin_proxy_path:
        return False

    return hutils.node.is_panel_active(domain.domain, child_admin_proxy_path)


def get_child_active_domains(child: Child) -> List[Domain]:
    actives = []
    for d in child.domains:  # type: ignore
        if is_child_domain_active(child, d):
            actives.append(d)
    return actives


def is_child_active(child: Child) -> bool:
    for d in child.domains:  # type: ignore
        if is_child_domain_active(child, d):
            return True
    return False




def generate_wireguard_config(proxy: dict) -> str:
    """
    Generates a WireGuard configuration from a given proxy dictionary.

    Args:
        proxy (dict): Dictionary containing WireGuard and proxy details.

    Returns:
        str: A WireGuard configuration string.
    """
    name=f'{proxy["extra_info"]} {proxy["name"]}'
    addrs = f"{proxy['wg_ipv4']}/32"
    if proxy['wg_ipv6']:
        addrs += f", {proxy['wg_ipv6']}/128"
    config = f"""[Interface]
# Name = {name}
Address= {addrs}
PrivateKey = {proxy["wg_pk"]}
MTU = {proxy.get("mtu", 1380)}
DNS = {proxy.get("dns", "1.1.1.1")}

[Peer]
# Name = Public Peer for {name}
Endpoint = {proxy["server"]}:{proxy["port"]}
PublicKey = {proxy["wg_server_pub"]}
PresharedKey = {proxy['wg_psk']}
#PersistentKeepalive = {proxy.get("keep_alive", 25)}
"""

#Address = {proxy.get("wg_ipv4", "0.0.0.0/32")}
#AllowedIPs = {proxy.get("allowed_ips", "0.0.0.0/0")}

    return config



import uuid
from apiflask.fields import String, Float, Enum, Date, Integer, Boolean, DateTime
from apiflask import Schema, fields
from typing import Any, Mapping

from marshmallow import ValidationError

from hiddifypanel.models import UserMode, Lang, AdminMode
from hiddifypanel import hutils

# region user api


class FriendlyDateTime(fields.Field):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs):
        return hutils.convert.time_to_json(value)

    def _deserialize(self, value: Any, attr: str | None, data: Mapping[str, Any] | None, **kwargs):
        return hutils.convert.json_to_time(value)


class FriendlyUUID(fields.Field):

    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs):
        if value is None or not hutils.auth.is_uuid_valid(value):
            return None
        return str(value)

    def _deserialize(self, value: Any, attr: str | None, data: Mapping[str, Any] | None, **kwargs):
        if value is None or not hutils.auth.is_uuid_valid(value):
            return None
        try:
            return str(uuid.UUID(value))
        except ValueError:
            self.fail('Invalid uuid')

    def _validated(self, value):
        if not hutils.auth.is_uuid_valid(value):
            raise ValidationError('Invalid UUID')


class UserSchema(Schema):
    uuid = FriendlyUUID(required=True,  metadata={"description": "Unique identifier for the user"})
    name = String(required=True,  metadata={"description": "Name of the user"})

    usage_limit_GB = Float(
        required=False,
        allow_none=True,
         metadata={"description": "The data usage limit for the user in gigabytes"}
    )
    package_days = Integer(
        required=False,
        allow_none=True,
         metadata={"description": "The number of days in the user's package"}
    )
    mode = Enum(UserMode,
                required=False,
                allow_none=True,
                 metadata={"description": "The mode of the user's account, which dictates access level or type"}
                )
    last_online = FriendlyDateTime(
        # format="%Y-%m-%d %H:%M:%S",
        allow_none=True,
         metadata={"description": "The last time the user was online, converted to a JSON-friendly format"}
    )
    start_date = Date(
        # format='%Y-%m-%d',
        allow_none=True,
         metadata={"description": "The start date of the user's package, in a JSON-friendly format"}
    )
    current_usage_GB = Float(
        required=False,
        allow_none=True,
         metadata={"description": "The current data usage of the user in gigabytes"}
    )
    last_reset_time = FriendlyDateTime(
        # format='%Y-%m-%d %H:%M:%S',
         metadata={"description": "The last time the user's data usage was reset, in a JSON-friendly format"},
        allow_none=True
    )
    # expiry_time = Date(
    #     format='%Y-%m-%d',
    #      metadata={"description": "The expiry time of the user's package, in a JSON-friendly format"},
    #     allow_none=True
    # )
    comment = String(
        load_default=None,
        allow_none=True,
         metadata={"description": "An optional comment about the user"}
    )
    added_by_uuid = FriendlyUUID(
        required=False,
         metadata={"description": "UUID of the admin who added this user"},
        allow_none=True,
        # validate=OneOf([p.uuid for p in AdminUser.query.all()])
    )
    telegram_id = Integer(
        required=False,
         metadata={"description": "The Telegram ID associated with the user"},
        allow_none=True
    )
    ed25519_private_key = String(
        required=False,
        allow_none=True,
         metadata={"description": "If empty, it will be created automatically, The user's private key using the Ed25519 algorithm"}
    )
    ed25519_public_key = String(
        required=False,
        allow_none=True,
         metadata={"description": "If empty, it will be created automatically,The user's public key using the Ed25519 algorithm"}
    )
    wg_pk = String(
        required=False,
        allow_none=True,
         metadata={"description": "If empty, it will be created automatically, The user's WireGuard private key"}
    )

    wg_pub = String(
        required=False,
        allow_none=True,
         metadata={"description": "If empty, it will be created automatically, The user's WireGuard public key"}
    )
    wg_psk = String(
        required=False,
        allow_none=True,
         metadata={"description": "If empty, it will be created automatically, The user's WireGuard preshared key"}
    )

    lang = Enum(Lang, required=False, allow_none=True,  metadata={"description": "The language of the user"})
    enable = Boolean(required=False,  metadata={"description": "Whether the user is enabled or not"})
    is_active = Boolean(required=False,  metadata={"description": "Whether the user is active for using hiddify"})
    id = Integer(required=False,  metadata={"description": "never use it, only for better presentation"})


class PostUserSchema(UserSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # the uuid is sent in the url path
        self.fields['uuid'].required = False
        self.fields['uuid'].allow_none = True
        del self.fields['id']


class PatchUserSchema(UserSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['uuid'].required = False
        self.fields['uuid'].allow_none = True,
        self.fields['name'].required = False
        self.fields['name'].allow_none = True,
        del self.fields['id']


# endregion

# region admin api


class AdminSchema(Schema):
    name = String(required=True,  metadata={"description": "The name of the admin"})
    comment = String(required=False,  metadata={"description": "A comment related to the admin"}, allow_none=True)
    uuid = FriendlyUUID(required=False, allow_none=True,  metadata={"description": "The unique identifier for the admin"})
    mode = Enum(AdminMode, required=True,  metadata={"description": "The mode for the admin"})
    can_add_admin = Boolean(required=True,  metadata={"description": "Whether the admin can add other admins"})
    parent_admin_uuid = FriendlyUUID(required=False,  metadata={"description": "The unique identifier for the parent admin"}, allow_none=True,
                                     # validate=OneOf([p.uuid for p in AdminUser.query.all()])
                                     )
    telegram_id = Integer(required=False,  metadata={"description": "The Telegram ID associated with the admin"}, allow_none=True)
    lang = Enum(Lang, required=True)
    max_users = Integer(required=False,  metadata={"description": "The maximum number of users allowed"}, allow_none=True)
    max_active_users = Integer(required=False,  metadata={"description": "The maximum number of active users allowed"}, allow_none=True)


class PatchAdminSchema(AdminSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['uuid'].required = False
        self.fields['name'].required = False
        self.fields['mode'].required = False
        self.fields['lang'].required = False
        self.fields['can_add_admin'].required = False

# endregion


class SuccessfulSchema(Schema):
    status = Integer()
    msg = String()



from apiflask import abort, fields, Schema
from flask.views import MethodView
from flask import current_app as app, g
from flask_babel import lazy_gettext as _
from loguru import logger

from hiddifypanel.auth import login_required
from hiddifypanel.models import set_hconfig, ConfigEnum, PanelMode, Role
from hiddifypanel import hutils

from .schema import RegisterWithParentInputSchema


class RegisterWithParentApi(MethodView):
    decorators = [login_required({Role.super_admin})]

    # TODO: incomplete (not used)
    @app.input(RegisterWithParentInputSchema, arg_name='data')  # type: ignore
    def post(self, data):
        logger.info(f"Registering panel with parent called by {data['parent_unique_id']}")
        if hutils.node.is_parent() or hutils.node.is_child():
            logger.error("The panel is not in standalone mode nor in child")
            abort(400, 'The panel is not in standalone mode nor in child')

        set_hconfig(ConfigEnum.parent_panel, data['parent_panel'])  # type: ignore

        if not hutils.node.child.register_to_parent(data['name'], data['apikey']):
            logger.error("Child registration to parent failed")
            set_hconfig(ConfigEnum.parent_panel, '')  # type: ignore
            abort(400, _('child.register-failed'))  # type: ignore

        set_hconfig(ConfigEnum.panel_mode, PanelMode.child)  # type: ignore
        logger.info("Registered panel with parent, panel mode is now child")
        return {'status': 200, 'msg': 'ok'}



import re
import uuid
from hiddifypanel import hutils
from hiddifypanel.models.role import Role
from hiddifypanel.panel import hiddify
from hiddifypanel.auth import login_required

from wtforms.validators import ValidationError

from hiddifypanel.models import ConfigEnum, Domain
from .adminlte import AdminLTEModelView
from flask import current_app


class ConfigAdmin(AdminLTEModelView):
    can_export = True
    column_default_sort = ('category', False)
    column_display_pk = True
    can_delete = False
    can_create = False
    column_editable_list = ["value"]
    column_searchable_list = ["key", "value"]
    form_widget_args = {
        'description': {
            'readonly': True
        },
    }

    def is_accessible(self):
        if login_required(roles={Role.super_admin})(lambda: True)() != True:
            return False
        return True

    def on_model_change(self, form, model, is_created):
        if model.key == ConfigEnum.db_version:
            raise ValidationError('DB version can not be changed')
        # if model.key==ConfigEnum.decoy_domain:
        #     if not re.match("http(s|)://([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})/?", model.value):
        #         raise ValidationError('Invalid address: e.g., https://www.wikipedia.org/')
        if model.key in [ConfigEnum.admin_secret]:
            if not hutils.encode.is_valid_uuid(model.value):
                raise ValidationError('Invalid UUID e.g.,' + str(uuid.uuid4()))

        # There is no telegram_secret !?
        # if model.key in [ConfigEnum.telegram_secret]:
        #     if not re.match("^[0-9a-fA-F]{32}$", model.value):
        #         raise ValidationError('Invalid UUID e.g.,' + uuid.uuid4().hex)

        if model.key == ConfigEnum.proxy_path:
            if not re.match("^[a-zA-Z0-9]*$", model.value):
                raise ValidationError('Invalid path. should be ascii string')

        if model.key in [ConfigEnum.tls_ports, ConfigEnum.kcp_ports, ConfigEnum.http_ports]:
            if not re.match("^(\\d,?)*$", model.value):
                raise ValidationError('Invalid path. should be comma separated integer e.g., 80,81')

        if model.key == ConfigEnum.http_ports:
            if "80" not in model.value.split(","):
                raise ValidationError('Port 80 should always be presented')
        if model.key == ConfigEnum.tls_ports:
            if "443" not in model.value.split(","):
                raise ValidationError('Port 443 should always be presented')

        if "domain" in model.key:
            if not re.match("^([A-Za-z0-9\\-.]+\\.[a-zA-Z]{2,})$", model.value):
                raise ValidationError('Invalid domain: e.g., www.google.com')
            if len(Domain.query.filter(Domain.domain == model.value).all()) > 0:
                raise ValidationError(f"Domain model.value is exist in domains section. Use a fake domain")



from enum import auto
import ipaddress
import re
from typing import Dict, List
from flask import request

from sqlalchemy.orm import backref
from strenum import StrEnum


from hiddifypanel.database import db
from hiddifypanel.models.config import hconfig
from .child import Child
from hiddifypanel.models.config_enum import ConfigEnum


class DomainType(StrEnum):
    direct = auto()
    sub_link_only = auto()
    cdn = auto()
    auto_cdn_ip = auto()
    relay = auto()
    worker = auto()
    fake = auto()

    reality = auto() #deprecated
    special_reality_tcp = auto()
    special_reality_xhttp = auto()
    special_reality_grpc = auto()
    old_xtls_direct = auto() #deprecated
    dnstt = auto()
    # special_shadowtls = auto()

    # fake_cdn = "fake_cdn"
    # telegram_faketls = "telegram_faketls"
    # ss_faketls = "ss_faketls"


ShowDomain = db.Table('show_domain',
                      db.Column('domain_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True),
                      db.Column('related_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True)
                      )



class Domain(db.Model):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0)
    domain = db.Column(db.String(200), nullable=True, unique=False)
    alias = db.Column(db.String(200))
    sub_link_only = db.Column(db.Boolean, nullable=False, default=False)
    mode = db.Column(db.Enum(DomainType), nullable=False, default=DomainType.direct)
    cdn_ip = db.Column(db.Text(2000), nullable=True, default='')
    # port_index=db.Column(db.Integer, nullable=True, default=0)
    grpc = db.Column(db.Boolean, nullable=True, default=False)
    servernames = db.Column(db.String(1000), nullable=True, default='')
    # show_all=db.Column(db.Boolean, nullable=True)
    show_domains = db.relationship('Domain', secondary=ShowDomain,
                                   primaryjoin=id == ShowDomain.c.domain_id,
                                   secondaryjoin=id == ShowDomain.c.related_id,
                                   backref=backref('showed_by_domains', lazy='dynamic')
                                   )
    download_domain_id= db.Column(db.Integer, db.ForeignKey('domain.id', ondelete='SET NULL'), default=None,nullable=True)
    download_domain = db.relationship('Domain',remote_side=[id],    foreign_keys=[download_domain_id])
    extra_params = db.Column(db.String(200), nullable=True, default='')
    resolve_ip= db.Column(db.Boolean, nullable=True, default=False)

    def __repr__(self):
        return f'{self.domain}'

    def get_cdn_ips_parsed(self):
        ips = re.split('[ \t\r\n;,]+', self.cdn_ip.strip())
        res = set()
        for ip in ips:
            try:
                res.add(ipaddress.ip_address(ip))
            except:
                pass
        return res

    def to_dict(self, dump_ports=False, dump_child_id=False):
        data = {
            'domain': self.domain.lower(),
            'mode': self.mode,
            'alias': self.alias,
            'sub_link_only': self.sub_link_only,
            'child_unique_id': self.child.unique_id if self.child else '',  # type: ignore
            'cdn_ip': self.cdn_ip,
            'servernames': self.servernames,
            'grpc': self.grpc,
            'download_domain':self.download_domain.domain if self.download_domain else "",
            'show_domains': [dd.domain for dd in self.show_domains],  # type: ignore
            "resolve_ip":self.resolve_ip,
        }
        if dump_child_id:
            data['child_id'] = self.child_id
        if dump_ports:
            data["internal_port_hysteria2"] = self.internal_port_hysteria2
            data["internal_port_tuic"] = self.internal_port_tuic
            data["internal_port_naive"] = self.internal_port_naive
            data["internal_port_special"] = self.internal_port_special
            data["internal_port_dnstt"] = self.internal_port_dnstt
            data["need_valid_ssl"] = self.need_valid_ssl

        return data

    @staticmethod
    def from_schema(schema):
        return schema.dump(Domain())

    def to_schema(self):
        domain_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import DomainSchema
        return DomainSchema().load(domain_dict)


    def auto_cdn_ip(self):
        from hiddifypanel import hutils
        if self.cdn_ip:
            return hutils.network.auto_ip_selector.get_clean_ip(self.cdn_ip)
        return None

    @property
    def need_valid_ssl(self):
        return self.mode in [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]

    @property
    def port_index(self):
        return self.id

    @property
    def internal_port_hysteria2(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        # print("child_id",self.child_id)
        return int(hconfig(ConfigEnum.hysteria_port, self.child_id)) + self.port_index

    @property
    def internal_port_dnstt(self):
        if self.mode not in [DomainType.dnstt]:
            return 0
        # TODO: check validity of the range of the port
        # print("child_id",self.child_id)
        return int(5400) + self.port_index


    @property
    def internal_port_tuic(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.tuic_port, self.child_id)) + self.port_index

    @property
    def internal_port_naive(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.naive_port, self.child_id)) + self.port_index

    @property
    def internal_port_special(self):
        if self.mode != DomainType.reality and "special" not in self.mode.value:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.special_port, self.child_id)) + self.port_index

    @classmethod
    def by_mode(cls, mode: DomainType) -> List['Domain']:
        domains = Domain.query.filter(Domain.mode == mode).all()
        if domains:
            return [d.domain for d in domains]
        return []

    @classmethod
    def modes_and_domains(cls) -> Dict[DomainType, List['Domain']]:
        return {mode: cls.by_mode(mode) for mode in DomainType}

    @classmethod
    def by_domain(cls, domain: str) -> 'Domain | None':
        return Domain.query.filter(Domain.domain == domain).first()

    @classmethod
    def get_panel_link(cls, child_id: int | None = None) -> str | None:
        if child_id is None:
            child_id = Child.current().id  # type: ignore
        domains = Domain.query.filter(Domain.mode.in_(
            [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]),
            Domain.child_id == child_id
        ).all()
        if not domains:
            return None
        return domains[0].domain

    @classmethod
    def get_domains(cls, always_add_ip=False, always_add_all_domains=False) -> List['Domain']:
        from hiddifypanel import hutils
        domains = []
        domains = db.session.query(Domain).filter(Domain.mode == DomainType.sub_link_only, Domain.child_id == Child.current().id).all()
        if not len(domains) or always_add_all_domains:
            domains = db.session.query(Domain).filter(Domain.mode.notin_([DomainType.fake, DomainType.reality,DomainType.special_reality_tcp,DomainType.special_reality_xhttp,DomainType.special_reality_grpc])).all()

        if len(domains) == 0 and request:
            domains = [Domain(domain=request.host)]  # type: ignore
        if len(domains) == 0 or always_add_ip:
            domains += [Domain(domain=hutils.network.get_ip_str(4))]  # type: ignore
        return domains

    @classmethod
    def add_or_update(cls, commit=True, child_id=0, **domain):
        dbdomain = Domain.query.filter(Domain.domain == domain['domain']).first()
        if not dbdomain:
            dbdomain = Domain(domain=domain['domain'])  # type: ignore
            db.session.add(dbdomain)
        dbdomain.child_id = child_id

        dbdomain.mode = domain['mode']
        if (str(domain.get('sub_link_only', False)).lower() == 'true'):
            dbdomain.mode = DomainType.sub_link_only
        dbdomain.cdn_ip = domain.get('cdn_ip', '')
        dbdomain.alias = domain.get('alias', '')
        dbdomain.grpc = domain.get('grpc', False)
        dbdomain.servernames = domain.get('servernames', '')
        dbdomain.resolve_ip=domain.get("resolve_ip",False)
        show_domains = domain.get('show_domains', [])
        dbdomain.show_domains = Domain.query.filter(Domain.domain.in_(show_domains)).all()
        dl_domain=domain.get("download_domain")
        if dl_domain:
            dbdldomain = Domain.query.filter(Domain.domain == dl_domain).first()
            if not dbdldomain:
                dbdldomain = Domain(domain=dl_domain)  # type: ignore
                db.session.add(dbdldomain)
                db.session.commit()
                dbdldomain=Domain.query.filter(Domain.domain == dl_domain).first()
            assert dbdldomain
            dbdomain.download_domain_id=dbdldomain.id
        if commit:
            db.session.commit()

    @classmethod
    def bulk_register(cls, domains, commit=True, remove=False, force_child_unique_id: str | None = None):
        from hiddifypanel.panel import hiddify
        child_ids = {}
        for domain in domains:
            child_id = hiddify.get_child(unique_id=force_child_unique_id)
            child_ids[child_id] = 1
            cls.add_or_update(commit=False, child_id=child_id, **domain)
        if remove and len(child_ids):
            dd = {d['domain']: 1 for d in domains}
            for d in Domain.query.filter(Domain.child_id.in_(child_ids)):
                if d.domain not in dd:
                    db.session.delete(d)

        if commit:
            db.session.commit()



from flask import render_template, request, g
import json

from hiddifypanel import hutils
from hiddifypanel.hutils.proxy.xrayjson import to_xray
from hiddifypanel.models import ProxyProto, ProxyTransport, Domain, ConfigEnum


def configs_as_json(domains: list[Domain], **kwargs) -> str:
    ua = hutils.flask.get_user_agent()
    base_config = json.loads(render_template('base_singbox_config.json.j2'))
    allphttp = [p for p in request.args.get("phttp", "").split(',') if p]
    allptls = [p for p in request.args.get("ptls", "").split(',') if p]

    allp = []
    for d in domains:
        base_config['dns']['rules'][0]['domain'].append(d.domain)
    endpoints=[]
    for pinfo in hutils.proxy.get_valid_proxies(domains):
        sing = to_singbox(pinfo)
        if 'msg' not in sing:
            if hutils.flask.is_client_version(hutils.flask.ClientVersion.hiddify_next, 4, 0, 0) and sing[0]['type']=="wireguard":
                endpoints+=sing
            else:
                allp += sing
    base_config['outbounds'] += allp
    base_config['endpoints'] = endpoints

    select = {
        "type": "selector",
        "tag": "Select",
        "outbounds": [p['tag'] for p in allp if 'shadowtls-out' not in p['tag']],
        "default": "Auto"
    }
    select['outbounds'].insert(0, "Auto")
    base_config['outbounds'].insert(0, select)
    smart = {
        "type": "urltest",
        "tag": "Auto",
        "outbounds": [p['tag'] for p in allp if 'shadowtls-out' not in p],
        "url": "https://www.gstatic.com/generate_204",
        "interval": "10m",
        "tolerance": 200
    }
    base_config['outbounds'].insert(1, smart)
    res = json.dumps(base_config, indent=4, cls=hutils.proxy.ProxyJsonEncoder)
    # if ua['is_hiddify']:
    #     res = res[:-1]+',"experimental": {}}'
    return res


def is_xray_proxy(proxy: dict):
    # if g.user_agent.get('is_hiddify_prefere_xray'):
    #     return True
    # if proxy['transport'] == ProxyTransport.xhttp:
    #     return True
    return False


def to_singbox(proxy: dict) -> list[dict] | dict:
    name = proxy['name']

    all_base = []
    if proxy['l3'] == "kcp":
        return {'name': name, 'msg': "clash does not support kcp", 'type': 'debug'}

    base = {}
    all_base.append(base)
    # vmess ws
    base["tag"] = f"""{proxy['extra_info']} {proxy["name"]} § {proxy['port']} {proxy["dbdomain"].id}"""
    if is_xray_proxy(proxy):
        if hutils.flask.is_client_version(hutils.flask.ClientVersion.hiddify_next, 1, 9, 0):
            base['type'] = "xray"
            xp = to_xray(proxy)
            xp['streamSettings']['sockopt'] = {}
            base['xray_outbound_raw'] = xp
            if proxy.get('tls_fragment_enable'):
                base['xray_fragment'] = {
                    'packets': proxy.get("tls_fragment_packets", "tlshello"),
                    'length': proxy["tls_fragment_size"],
                    'interval': proxy["tls_fragment_sleep"]
                }
            return all_base
        return {'name': name, 'msg': "xray proxy does not support in this client version", 'type': 'debug'}
    base["type"] = str(proxy["proto"])
    if proxy['proto']==ProxyProto.dnstt:
        add_dnstt(all_base, proxy)
        return all_base
    
    base["server"] = proxy["server"]
    base["server_port"] = int(proxy["port"])
    # base['alpn'] = proxy['alpn'].split(',')
    if proxy["proto"] == "ssr":
        add_ssr(base, proxy)
        return all_base
    if proxy["proto"] == ProxyProto.wireguard:
        add_wireguard(base, proxy)
        return all_base
    
    
    if proxy['proto']==ProxyProto.mieru:
        add_mieru(base, proxy)
        return all_base
    if proxy['proto']==ProxyProto.naive:
        add_naive(base, proxy)
        return all_base
    if proxy["proto"] in ["ss", "v2ray"]:
        add_shadowsocks_base(all_base, proxy)
        return all_base
    if proxy["proto"] == "ssh":
        add_ssh(all_base, proxy)
        return all_base

    if proxy["proto"] == "trojan":
        base["password"] = proxy["password"]

    if proxy['proto'] in ['vmess', 'vless']:
        base["uuid"] = proxy["uuid"]

    if proxy['proto'] in ['vmess', 'vless', 'trojan']:
        add_multiplex(base, proxy)

    add_tls(base, proxy)

    if g.user_agent.get('is_hiddify'):
        add_tls_tricks(base, proxy)

    if proxy.get('flow'):
        base["flow"] = proxy['flow']
        # base["flow-show"] = True

    if proxy["proto"] == "vmess":
        base["alter_id"] = 0
        base["security"] = proxy["cipher"]

    # base["udp"] = True
    if proxy["proto"] in ["vmess", "vless"]:
        base["packet_encoding"] = "xudp"  # udp packet encoding

    if proxy["proto"] == "tuic":
        add_tuic(base, proxy)
    elif proxy["proto"] == "hysteria2":
        add_hysteria(base, proxy)
    else:
        add_transport(base, proxy)

    return all_base


def add_multiplex(base: dict, proxy: dict):
    if proxy.get('mux_enable') != "singbox":
        return
    base['multiplex'] = {
        "enabled": True,
        "protocol": proxy['mux_protocol'],
        "padding": proxy['mux_padding_enable']
    }
    # Conflicts: max_streams with max_connections and min_streams
    mux_max_streams = proxy.get('mux_max_streams', 0)
    if mux_max_streams and mux_max_streams != 0:
        base['multiplex']['max_streams'] = mux_max_streams
    else:
        base['multiplex']['max_connections'] = proxy.get('mux_max_connections', 0)
        base['multiplex']['min_streams'] = proxy.get('mux_min_streams', 0)

    add_tcp_brutal(base, proxy)


def add_tcp_brutal(base: dict, proxy: dict):
    if 'multiplex' in base:
        if proxy.get('mux_brutal_enable'):
            base['multiplex']['brutal'] = {
                "enabled": proxy.get('mux_brutal_enable', False),
                "up_mbps": proxy.get('mux_brutal_up_mbps', 10),
                "down_mbps": proxy.get('mux_brutal_down_mbps', 10)
            }


def add_udp_over_tcp(base: dict):
    base['udp_over_tcp'] = {
        "enabled": True,
        "version": 2
    }


def add_tls(base: dict, proxy: dict):
    if not ("tls" in proxy["l3"] or "reality" in proxy["l3"] or "quic" in proxy["l3"]):
        return
    base["tls"] = {
        "enabled": True,
        "server_name": proxy["sni"]
    }
    if proxy.get("ech"):
        base["tls"]['ech'] = {
            "enabled": True,
            "config":f"-----BEGIN ECH CONFIGS-----\\n{proxy.get("ech")}\\n-----END ECH CONFIGS-----"
        }   
    if proxy['proto']=="naive":
        return
    if proxy['proto'] not in ["tuic", "hysteria2"] and proxy['transport']!="xhttp":
        base["tls"]["utls"] = {
            "enabled": True,
            "fingerprint": proxy.get('fingerprint', 'none')
        }

    if "reality" in proxy["l3"]:
        base["tls"]["utls"] = {
            "enabled": True,
            "fingerprint": proxy.get('fingerprint', 'none')
        }
        base["tls"]["reality"] = {
            "enabled": True,
            "public_key": proxy['reality_pbk'],
            "short_id": proxy['reality_short_id']
        }
    base["tls"]['insecure'] = proxy['allow_insecure'] or (proxy["mode"] == "Fake")
    
    base["tls"]["alpn"] = proxy['alpn'].split(',')
    


def add_tls_tricks(base: dict, proxy: dict):
    if proxy.get('tls_fragment_enable'):
        base['tls_fragment'] = {
            'enabled': True,
            'size': proxy["tls_fragment_size"],
            'sleep': proxy["tls_fragment_sleep"]
        }

    if 'tls' in base:
        if proxy.get("tls_padding_enable") or proxy.get("tls_mixed_case"):
            base['tls']['tls_tricks'] = {}
        if proxy.get("tls_padding_enable"):
            base['tls']['tls_tricks']['padding_size'] = proxy["tls_padding_length"]

        if proxy.get("tls_mixed_case"):
            base['tls']['tls_tricks']['mixedcase_sni'] = True


def add_transport(base: dict, proxy: dict):
    if proxy['l3'] == 'reality' and proxy['transport'] not in ["grpc"]:
        return
    base["transport"] = {}
    if proxy['transport'] in ["ws", "WS"]:
        base["transport"] = {
            "type": "ws",
            "path": proxy["path"],
            "early_data_header_name": "Sec-WebSocket-Protocol"
        }
        if "host" in proxy:
            base["transport"]["headers"] = {"Host": proxy["host"]}

    if proxy['transport'] in ["xhttp"]:
        _add_xhttp_details(base,proxy)
        

    if proxy['transport'] in [ProxyTransport.httpupgrade]:
        base["transport"] = {
            "type": "httpupgrade",
            "path": proxy["path"]
        }
        if "host" in proxy:
            base["transport"]["headers"] = {"Host": proxy["host"]}

    if proxy["transport"] in ["tcp", "h2"]:
        base["transport"] = {
            "type": "http",
            "path": proxy.get("path", ""),
            "idle_timeout": "15s",
            "ping_timeout": "15s"
            # "method": "",
            # "headers": {},
        }

        if 'host' in proxy:
            base["transport"]["host"] = [proxy["host"]]

    if proxy["transport"] == "grpc":
        base["transport"] = {
            "type": "grpc",
            "service_name": proxy["grpc_service_name"],
            "idle_timeout": "115s",
            "ping_timeout": "15s",
            # "permit_without_stream": false
        }

def _add_xhttp_details(base: dict, proxy: dict):
    
    base["transport"] = {
            "type": "xhttp",
            "path": proxy["path"],
            'host': proxy['host'],
            'mode':proxy['xhttp_mode'],
            "headers": proxy['params'].get('headers', {})
        }
    
    
    if pdl:=proxy.get("download"):
        base['transport']['downloadSettings']={
            "path": proxy["path"],
            'host': pdl.get("server"),
            "headers":pdl.get("headers")            
        }
        dls={
            'l3':proxy['l3'],
            'proto':proxy['proto'],
            "transport":proxy['transport'],
            **proxy['download']
        }
        add_tls(base['transport']['downloadSettings'],dls)
        

def add_dnstt(all_base:list,proxy:dict):
    all_base[0]['domain']=proxy["sni"]
    all_base[0]['publicKey']=proxy["public_key"]
    all_base[0]['resolvers']=proxy['resolvers']
    all_base[0]['tunnel_per_resolver']=proxy['tunnel_per_resolver']
    tag=all_base[0]["tag"]
    all_base[0]["tag"]+="§hide§"
    all_base.append({
        "type":"socks",
        "username":proxy['uuid'],
        "password":proxy['password'],
        "tag":tag,
        "detour":all_base[0]["tag"]
    })
    
 

def add_ssr(base: dict, proxy: dict):

    base["method"] = proxy["cipher"]
    base["password"] = proxy["uuid"]
    # base["udp"] = True
    base["obfs"] = proxy["ssr-obfs"]
    base["protocol"] = proxy["ssr-protocol"]
    base["protocol-param"] = proxy["fakedomain"]


def add_naive(base: dict, proxy: dict):
    base['type']="naive"
    base['username']=proxy['uuid']
    base["password"]=proxy["password"]
    base["udp_over_tcp"]=True
    base["quic"]=proxy["quic"]
    base["extra_headers"]={
        "hiddify-naive-secret":proxy["path"]
    }

    add_tls(base,proxy)

def add_mieru(base: dict, proxy: dict):
    base['type']="mieru"
    base['multiplexing']=proxy['multiplexing']
    base['handshake_mode']=proxy['handshake']
    base['username']=proxy['uuid']
    base['password']=proxy['password']
    base['portBindings']=[]
    
    for port in proxy["tcp_ports"]:
        if port:
            base['portBindings'].append({
                'protocol':"TCP",
                "port":0 if "-" in port else int(port),
                "portRange":port if "-" in port else ""
            })
    for port in proxy["udp_ports"]:
        if port:
            base['portBindings'].append({
                'protocol':"UDP",
                "port":0 if "-" in port else int(port),
                "portRange":port if "-" in port else ""
            })
            
                
    
def add_wireguard(base: dict, proxy: dict):
    if hutils.flask.is_client_version(hutils.flask.ClientVersion.singbox, 1, 13, 0):
        
        base["private_key"] = proxy["wg_pk"]
        base["mtu"] = 1380
        base['peers']=[{
            "public_key":proxy["wg_server_pub"],
            "pre_shared_key":proxy["wg_psk"],
            "address":base['server'],
            "port":base['server_port'],
            "allowed_ips": [
                "0.0.0.0/0","::/0"
            ]
            # "address" : f'{proxy["wg_ipv4"]}/32'
        }]
        del base["server_port"]
        del base["server"]
        if g.user_agent.get('is_hiddify'):
            base["noise"] ={
                "fake_packet":{
                    "enabled":True,
                    "count":proxy["wg_noise_trick"]
                } 
            }

    else:
        base["local_address"] = f'{proxy["wg_ipv4"]}/32'
        base["private_key"] = proxy["wg_pk"]
        base["peer_public_key"] = proxy["wg_server_pub"]

        base["pre_shared_key"] = proxy["wg_psk"]

        base["mtu"] = 1380
        if g.user_agent.get('is_hiddify') and hutils.flask.is_client_version(hutils.flask.ClientVersion.hiddify_next, 0, 15, 0):
                base["fake_packets"] = proxy["wg_noise_trick"]


def add_shadowsocks_base(all_base: list[dict], proxy: dict):
    base = all_base[0]
    base["type"] = "shadowsocks"
    base["method"] = proxy["cipher"]
    base["password"] = proxy["password"]
    add_udp_over_tcp(base)
    add_multiplex(base, proxy)
    if proxy["transport"] == "faketls":
        base["plugin"] = "obfs-local"
        base["plugin_opts"] = f'obfs=tls;obfs-host={proxy["fakedomain"]}'
    if proxy['proto'] == 'v2ray':
        base["plugin"] = "v2ray-plugin"
        # "skip-cert-verify": proxy["mode"] == "Fake" or proxy['allow_insecure'],
        base["plugin_opts"] = f'mode=websocket;path={proxy["path"]};host={proxy["host"]};tls'

    if proxy["transport"] == "shadowtls":
        base['detour'] = base['tag'] + "_shadowtls-out §hide§"

        shadowtls_base = {
            "type": "shadowtls",
            "tag": base['detour'],
            "server": base['server'],
            "server_port": base['server_port'],
            "version": 3,
            "password": proxy["shared_secret"],
            "tls": {
                "enabled": True,
                "server_name": proxy["fakedomain"],
                "utls": {
                    "enabled": True,
                    "fingerprint": proxy.get('fingerprint', 'none')
                },
                # "alpn": proxy['alpn'].split(',')
            }
        }
        # add_utls(shadowtls_base)
        del base['server']
        del base['server_port']
        all_base.append(shadowtls_base)


def add_ssh(all_base: list[dict], proxy: dict):
    base = all_base[0]
    # base["client_version"]= "{{ssh_client_version}}"
    base["user"] = proxy['uuid']
    base["private_key"] = proxy['private_key']  # .replace('\n', '\\n')
    base["udp_over_tcp"]= True
    base["host_key"] = proxy.get('host_keys', [])


def add_tuic(base: dict, proxy: dict):
    base['congestion_control'] = "cubic"
    base['udp_relay_mode'] = 'native'
    base['zero_rtt_handshake'] = True
    base['heartbeat'] = "10s"
    base['password'] = proxy['uuid']
    base['uuid'] = proxy['uuid']


def add_hysteria(base: dict, proxy: dict):
    base['up_mbps'] = proxy.get(ConfigEnum.hysteria_up_mbps)
    base['down_mbps'] = proxy.get(ConfigEnum.hysteria_down_mbps)
    # TODO: check the obfs should be empty or not exists at all
    if proxy.get('hysteria_obfs_enable'):
        base['obfs'] = {
            "type": "salamander",
            "password": proxy.get('hysteria_obfs_password')
        }
    base['password'] = proxy['uuid']




from flask.views import MethodView
from flask import current_app as app
from flask import g
from apiflask import abort

from hiddifypanel.models.user import User
from hiddifypanel.database import db
from hiddifypanel.models.child import Child
from loguru import logger
from hiddifypanel.models import *
from hiddifypanel.auth import login_required
from .schema import SyncInputSchema, SyncOutputSchema


class SyncApi(MethodView):
    decorators = [login_required(node_auth=True)]

    @app.input(SyncInputSchema, arg_name='data')  # type: ignore
    @app.output(SyncOutputSchema)  # type: ignore
    def put(self, data):
        from hiddifypanel import hutils
        unique_id = Child.node.unique_id

        logger.info(f"Sync child with unique_id: {unique_id}")
        if not hutils.node.is_parent():
            logger.error("Not a parent")
            abort(400, "Not a parent")

        child = Child.query.filter(Child.unique_id == unique_id).first()
        if not child:
            logger.error("The child does not exist")
            abort(404, "The child does not exist")

        try:
            logger.info("Syncing domains...")
            if data.get('domains'):
                logger.info("Inserting domains into database")
                Domain.bulk_register(data['domains'], commit=False, force_child_unique_id=child.unique_id)
            else:
                logger.info("Domains field is empty")

            logger.info("Syncing hconfigs...")
            if data.get('hconfigs'):
                logger.info("Inserting hconfigs into database")
                bulk_register_configs(data['hconfigs'], commit=False, froce_child_unique_id=child.unique_id)
            else:
                logger.info("Hconfigs field is empty")

            logger.info("Syncing proxies...")
            if data.get('proxies'):
                logger.info("Inserting proxies into database")
                Proxy.bulk_register(data['proxies'], commit=False, force_child_unique_id=child.unique_id)
            else:
                logger.info("Proxies field is empty")

            logger.info("Commit changes to database")
            db.session.commit()  # type: ignore
        except Exception as err:
            with logger.contextualize(error=err):
                logger.error(f"Error while syncing data")
            abort(400, str(err))

        res = SyncOutputSchema()
        res.users = [u.to_schema() for u in User.query.all()]  # type: ignore
        res.admin_users = [a.to_schema() for a in AdminUser.query.all()]  # type: ignore

        logger.info("Returning sync output")
        return res



from apiflask import APIBlueprint

bp = APIBlueprint("api_parent", __name__, url_prefix="/<proxy_path>/api/v2/parent/", enable_openapi=False)


def init_app(app):
    with app.app_context():
        from .register_api import RegisterApi
        from .sync_api import SyncApi
        from .usage_api import UsageApi
        from .status_api import StatusApi
        bp.add_url_rule('/status/', view_func=StatusApi)
        bp.add_url_rule('/register/', view_func=RegisterApi)
        bp.add_url_rule('/sync/', view_func=SyncApi)
        bp.add_url_rule('/usage/', view_func=UsageApi)
    app.register_blueprint(bp)



from apiflask import APIBlueprint
from flask import g
from hiddifypanel.models import AdminUser, User

bp = APIBlueprint("api_admin", __name__, url_prefix="/<proxy_path>/api/v2/admin/", enable_openapi=True)


def init_app(app):

    with app.app_context():
        from .admin_info_api import AdminInfoApi
        from .server_status_api import AdminServerStatusApi
        from .admin_user_api import AdminUserApi
        from .admin_users_api import AdminUsersApi
        from .admin_log_api import AdminLogApi
        from .system_actions import UpdateUserUsageApi, AllConfigsApi
        bp.add_url_rule('/me/', view_func=AdminInfoApi)  # type: ignore
        bp.add_url_rule('/server_status/', view_func=AdminServerStatusApi)  # type: ignore
        bp.add_url_rule('/admin_user/<uuid:uuid>/', view_func=AdminUserApi)  # type: ignore
        bp.add_url_rule('/admin_user/', view_func=AdminUsersApi)  # type: ignore
        bp.add_url_rule('/log/', view_func=AdminLogApi)  # type: ignore
        bp.add_url_rule('/update_user_usage/', view_func=UpdateUserUsageApi)  # type: ignore
        bp.add_url_rule('/all-configs/', view_func=AllConfigsApi)  # type: ignore
        from .user_api import UserApi
        from .users_api import UsersApi
        bp.add_url_rule('/user/<uuid:uuid>/', view_func=UserApi)  # type: ignore
        bp.add_url_rule('/user/', view_func=UsersApi)  # type: ignore
    app.register_blueprint(bp)


def has_permission(model) -> bool:
    '''Check if the authenticated account has permission to do an action(get,insert,update,delete) on the another admin'''
    if g.account.uuid == AdminUser.get_super_admin_uuid():
        return True
    if isinstance(model, AdminUser) and model.parent_admin_id == g.account.id:
        return True
    elif isinstance(model, User) and model.added_by == g.account.id:
        return True

    return False



from apiflask import abort
from flask.views import MethodView
from flask import current_app as app
from flask import g
from loguru import logger

from hiddifypanel.models import Child
from hiddifypanel.panel.usage import add_users_usage_uuid
from hiddifypanel.auth import login_required

from .schema import UsageInputOutputSchema


class UsageApi(MethodView):
    decorators = [login_required(node_auth=True)]

    @app.input(UsageInputOutputSchema, arg_name='data')  # type: ignore
    @app.output(UsageInputOutputSchema)  # type: ignore
    def put(self, data):
        from hiddifypanel import hutils
        child = Child.query.filter(Child.unique_id == Child.node.unique_id).first()
        if not child:
            logger.error("The child does not exist")
            abort(400, "The child does not exist")

        # parse request data
        logger.debug(f"Received Usage data from child: {data}")
        child_usages_data = hutils.node.convert_usage_api_response_to_dict(data)

        # get current usage
        logger.debug("Getting current usage data from parent")
        parent_current_usages_data = hutils.node.convert_usage_api_response_to_dict(UsageInputOutputSchema().dump(hutils.node.get_users_usage_data_for_api()))  # type: ignore

        # calculate usages
        logger.debug("Calculating increased usages")
        increased_usages = self.__calculate_parent_increased_usages(child_usages_data, parent_current_usages_data)
        logger.debug(f"Increased usages: {increased_usages}")

        # add users usage
        if increased_usages:
            logger.info(f"Adding increased usages to parent: {increased_usages}")
            add_users_usage_uuid(increased_usages, child.id)

        return hutils.node.get_users_usage_data_for_api()

    def __calculate_parent_increased_usages(self, child_usages_data: dict, parent_usages_data: dict) -> dict:
        res = {}
        for p_uuid, p_usage in parent_usages_data.items():
            if child_usage := child_usages_data.get(p_uuid):
                if child_usage['usage'] > 0:
                    usage_data = {
                        'usage':  child_usage['usage'] - p_usage['usage'],
                        'devices': child_usage['devices'],
                    }
                    if usage_data['usage'] > 0:
                        res[p_uuid] = usage_data
        return res



from .VERSION import __version__, __release_time__,is_released_version

from dotenv import load_dotenv
load_dotenv("/opt/hiddify-manager/hiddify-panel/app.cfg")

# from . import cache
from . import Events
from .base import create_app, create_app_wsgi
# from . import hutils
# from . import panel


__all__ = ["create_app", "create_app_wsgi"]

# application = create_app_wsgi()



from apiflask import Schema, fields


class RegisterWithParentInputSchema(Schema):
    parent_panel = fields.String(required=True,  metadata={"description": "The parent panel url"})
    name = fields.String(required=True,  metadata={"description": "The child's name in the parent panel"})
    apikey = fields.String( metadata={"description": "The parent's apikey"})
